/*****************************************************************************/
/**
 *  @file   TFE_main.cpp
 *  @author Naohisa Sakamoto
 */
/*----------------------------------------------------------------------------
 *
 *  Copyright (c) Visualization Laboratory, Kyoto University.
 *  All rights reserved.
 *  See http://www.viz.media.kyoto-u.ac.jp/kvs/copyright/ for details.
 *
 *  $Id: TFE_main.cpp 631 2010-10-10 02:15:35Z naohisa.sakamoto $
 */
/*****************************************************************************/
#include "TFE_main.h"
#include <kvs/DebugNew>
#include <kvs/MouseEvent>
#include <kvs/Math>
#include <kvs/HSVColor>
#include <kvs/RGBColor>
#include <kvs/Date>
#include <kvs/Time>
#include <kvs/glut/GLUT>
#include <kvs/IgnoreUnusedVariable>
#include <SupportGLUT/Viewer/KVSKey.h>
#include <SupportGLUT/Viewer/KVSMouseButton.h>

#include <fstream>
#include <algorithm>
#include "ParamInfo.h"

//#define _NO_HISTOGRAM

extern void cbTFEpanel_sw( int i );

namespace
{

class ResetButton : public kvs::glut::PushButton
{
    kvs::visclient::TFE_main* m_editor;

public:

    ResetButton( kvs::visclient::TFE_main* editor ):
        kvs::glut::PushButton( editor ),
        m_editor( editor )
    {
    }

    void released( void )
    {
        m_editor->reset();
    }
};

class ApplyButton : public kvs::glut::PushButton
{
    kvs::visclient::TFE_main* m_editor;

public:

    ApplyButton( kvs::visclient::TFE_main* editor ):
        kvs::glut::PushButton( editor ),
        m_editor( editor )
    {
    }

    void released( void )
    {
        m_editor->apply();
    }
};

} // end of namespace

namespace
{
kvs::visclient::TFE_main* ptfe;

void cb_resolution( int id )
{
    ptfe->resolution();
}

void cb_function_name( int id )
{
    ptfe->function_name();
}

void cb_function_variable( int id )
{
    ptfe->function_variable();
}

void cb_import_file( int id )
{
    ptfe->import_file();
}

void cb_export_file( int id )
{
    ptfe->export_file();
}

void cb_show_tfe_c1( int id )
{
    ptfe->show_tfe_c1();
}

void cb_show_tfe_c2( int id )
{
    ptfe->show_tfe_c2();
}

void cb_show_tfe_c3( int id )
{
    ptfe->show_tfe_c3();
}

void cb_show_tfe_c4( int id )
{
    ptfe->show_tfe_c4();
}

void cb_show_tfe_o1( int id )
{
    ptfe->show_tfe_o1();
}

void cb_show_tfe_o2( int id )
{
    ptfe->show_tfe_o2();
}

void cb_show_tfe_o3( int id )
{
    ptfe->show_tfe_o3();
}

} // end of namespace

namespace kvs
{

namespace visclient
{


TFE_main::TFE_main( Command* command, kvs::ScreenBase* parent ):
    m_command( command ),
    m_screen( parent ),
    m_color_map_palette( NULL ),
    m_opacity_map_palette( NULL ),
    m_color_histogram( NULL ),
    m_opacity_histogram( NULL ),
    m_reset_button( NULL ),
    m_apply_button( NULL ),
    m_min_value( 0.0f ),
    m_max_value( 1.0f ),
    m_current_tf( 0 )
{
    const std::string title = "Transfer Function Editor";
    const int x = ( parent != 0 ) ? parent->x() + parent->width() + 5 : 0;
    const int y = ( parent != 0 ) ? parent->y() : 0;
    const int width = 700;
#ifndef _NO_HISTOGRAM
    const int height = 640;
#else
    const int height = 550;
#endif
    const int margin = 10;
    const kvs::RGBColor color( 200, 200, 200 );

    SuperClass::background()->setColor( color );
    SuperClass::setTitle( title );
    SuperClass::setPosition( x, y );
    SuperClass::setSize( width, height );

    m_command->extransfuncedit = this;;

    m_color_map_palette = new kvs::glut::ColorMapPalette( this );
    m_color_map_palette->setCaption( "Color" );
    m_color_map_palette->setColorMap( m_doc.transferFunction[0].colorMap() );
//  m_color_map_palette->setX( m_color_palette->x0() );
//  m_color_map_palette->setY( m_color_palette->y1() - m_color_palette->margin() );
    m_color_map_palette->setY( 300 );
    m_color_map_palette->setHeight( 100 );
    m_color_map_palette->setEventType(
        kvs::EventBase::PaintEvent |
        kvs::EventBase::ResizeEvent );
    m_color_map_palette->show();

    m_opacity_map_palette = new kvs::glut::OpacityMapPalette( this );
    m_opacity_map_palette->setCaption( "Opacity" );
    m_opacity_map_palette->setOpacityMap( m_doc.transferFunction[0].opacityMap() );
    m_opacity_map_palette->setX( m_color_map_palette->x1() );
    m_opacity_map_palette->setY( m_color_map_palette->y0() );
    m_opacity_map_palette->setHeight( 100 );
    m_opacity_map_palette->setEventType(
        kvs::EventBase::PaintEvent |
        kvs::EventBase::ResizeEvent );
    m_opacity_map_palette->show();

#ifndef _NO_HISTOGRAM
    m_color_histogram = new kvs::visclient::Histogram( this );
    m_color_histogram->setCaption( "Histogram" );
    m_color_histogram->setX( m_color_map_palette->x0() );
    m_color_histogram->setY( m_color_map_palette->y1() - m_opacity_map_palette->margin() );
    m_color_histogram->setHeight( 100 );
    m_color_histogram->setGraphColor( kvs::RGBAColor( 100, 100, 100, 1.0f ) );

    m_opacity_histogram = new kvs::visclient::Histogram( this );
    m_opacity_histogram->setCaption( "Histogram" );
    m_opacity_histogram->setX( m_color_map_palette->x1() );
    m_opacity_histogram->setY( m_color_histogram->y0() );
    m_opacity_histogram->setHeight( 100 );
    m_opacity_histogram->setGraphColor( kvs::RGBAColor( 100, 100, 100, 1.0f ) );

    std::vector<size_t> table( 1024 );
    m_color_histogram->setTable( kvs::visclient::FrequencyTable( 0.0, 1.0, table.size(), table.data() ) );
    m_opacity_histogram->setTable( kvs::visclient::FrequencyTable( 0.0, 1.0, table.size(), table.data() ) );

    m_color_histogram->show();
    m_opacity_histogram->show();
#endif

    const size_t button_margin = 5;
    const size_t button_width = ( width - 2 * margin - button_margin ) / 2;

    m_reset_button = new ::ResetButton( this );
    m_reset_button->setCaption( "Reset" );
#ifndef _NO_HISTOGRAM
    m_reset_button->setX( m_color_histogram->x0() + m_color_histogram->margin() );
    m_reset_button->setY( m_color_histogram->y1() + m_color_histogram->margin() );
#else
    m_reset_button->setX( m_color_map_palette->x0() + m_color_map_palette->margin() );
    m_reset_button->setY( m_color_map_palette->y1() + m_color_map_palette->margin() );
#endif
    m_reset_button->setWidth( button_width );
    m_reset_button->show();

    m_apply_button = new ::ApplyButton( this );
    m_apply_button->setCaption( "Apply" );
    m_apply_button->setX( m_reset_button->x1() + button_margin );
    m_apply_button->setY( m_reset_button->y0() );
    m_apply_button->setWidth( ( width - margin ) / 2 - m_opacity_map_palette->margin() );
    m_apply_button->show();

    m_tfe_c1 = new kvs::visclient::TFE_Color_1( this );
    m_tfe_c2 = new kvs::visclient::TFE_Color_2( this );
    m_tfe_c3 = new kvs::visclient::TFE_Color_3( this );
    m_tfe_c4 = new kvs::visclient::TFE_Color_4( this );
    m_tfe_o1 = new kvs::visclient::TFE_Opacity_1( this );
    m_tfe_o2 = new kvs::visclient::TFE_Opacity_2( this );
    m_tfe_o3 = new kvs::visclient::TFE_Opacity_3( this );
}

TFE_main::~TFE_main( void )
{
    if ( m_color_map_palette ) delete m_color_map_palette;
    if ( m_opacity_map_palette ) delete m_opacity_map_palette;
    if ( m_color_histogram ) delete m_color_histogram;
    if ( m_opacity_histogram ) delete m_opacity_histogram;
    if ( m_reset_button ) delete m_reset_button;
    if ( m_apply_button ) delete m_apply_button;
}

int TFE_main::show( void )
{
    int id = SuperClass::show();
    glutSetWindow( m_id );
    ::ptfe = this;

    m_glui_common = GLUI_Master.create_glui_subwindow( m_id, GLUI_SUBWINDOW_TOP );
    m_glui_tferng = GLUI_Master.create_glui_subwindow( m_id, GLUI_SUBWINDOW_TOP );
    m_glui_tffctl = GLUI_Master.create_glui_subwindow( m_id, GLUI_SUBWINDOW_BOTTOM );
    m_glui_tffile = GLUI_Master.create_glui_subwindow( m_id, GLUI_SUBWINDOW_BOTTOM );

    m_glui_common->set_main_gfx_window( m_id );
    m_glui_tferng->set_main_gfx_window( m_id );
    m_glui_tffctl->set_main_gfx_window( m_id );
    m_glui_tffile->set_main_gfx_window( m_id );

    m_glui_resolution = m_glui_common->add_edittext( "Transfer Function RESOLUTION", GLUI_EDITTEXT_INT, NULL, -1, cb_resolution );
    m_glui_tf_synth   = m_glui_common->add_edittext( "Transfer Function SYNTHESIZER", GLUI_EDITTEXT_TEXT, NULL, -1, cb_function_variable );
    m_glui_tf_name    = m_glui_common->add_listbox( "Transfer Function Name", NULL, -1, cb_function_name );
    m_glui_resolution->set_w( 250 );
    m_glui_tf_synth->set_w( 600 );
    m_glui_tf_name->set_w( 250 );

    for ( size_t i = 0; i < m_doc.transferFunction.size(); i++ )
    {
        m_glui_tf_name->add_item( i, m_doc.transferFunction[i].Name.c_str() );
    }

    m_glui_tferng->add_statictext( "Transfer Function Color Map:" );
    m_glui_tf_var_c = m_glui_tferng->add_edittext( "Variable", GLUI_EDITTEXT_TEXT, NULL, -1, cb_function_variable );
    m_glui_tf_min_c = m_glui_tferng->add_edittext( "Range Min.", GLUI_EDITTEXT_FLOAT, NULL, -1, cb_function_variable );
    m_glui_tf_max_c = m_glui_tferng->add_edittext( "Range Max.", GLUI_EDITTEXT_FLOAT, NULL, -1, cb_function_variable );
    m_glui_tf_var_c->set_w( 240 );
    m_glui_tf_min_c->set_w( 140 );
    m_glui_tf_max_c->set_w( 140 );

    m_glui_rng_min_c = m_glui_tferng->add_statictext( "Server side range min: xxx" );
    m_glui_rng_max_c = m_glui_tferng->add_statictext( "Server side range max: xxx" );

    GLUI_Button* c1 = m_glui_tferng->add_button( "Color Map Editor (freeform curve)", id, cb_show_tfe_c1 );
    GLUI_Button* c2 = m_glui_tferng->add_button( "Color Map Editor (expression)", id, cb_show_tfe_c2 );
    GLUI_Button* c3 = m_glui_tferng->add_button( "Color Map Editor (control points)", id, cb_show_tfe_c3 );
    GLUI_Button* c4 = m_glui_tferng->add_button( "Color Map Editor (select colormap)", id, cb_show_tfe_c4 );
    c1->set_w( 340 );
    c2->set_w( 340 );
    c3->set_w( 340 );
    c4->set_w( 340 );

    m_glui_tferng->add_column( false );

    m_glui_tferng->add_statictext( "Transfer Function Opacity Map:" );
    m_glui_tf_var_o = m_glui_tferng->add_edittext( "Variable", GLUI_EDITTEXT_TEXT, NULL, -1, cb_function_variable );
    m_glui_tf_min_o = m_glui_tferng->add_edittext( "Range Min.", GLUI_EDITTEXT_FLOAT, NULL, -1, cb_function_variable );
    m_glui_tf_max_o = m_glui_tferng->add_edittext( "Range Max.", GLUI_EDITTEXT_FLOAT, NULL, -1, cb_function_variable );
    m_glui_tf_var_o->set_w( 240 );
    m_glui_tf_min_o->set_w( 140 );
    m_glui_tf_max_o->set_w( 140 );

    m_glui_rng_min_o = m_glui_tferng->add_statictext( "Server side range min: xxx" );
    m_glui_rng_max_o = m_glui_tferng->add_statictext( "Server side range max: xxx" );

    GLUI_Button* o1 = m_glui_tferng->add_button( "Opacity Map Editor (freeform curve)", id, cb_show_tfe_o1 );
    GLUI_Button* o2 = m_glui_tferng->add_button( "Opacity Map Editor (expression)", id, cb_show_tfe_o2 );
    GLUI_Button* o3 = m_glui_tferng->add_button( "Opacity Map Editor (control points)", id, cb_show_tfe_o3 );
    o1->set_w( 340 );
    o2->set_w( 340 );
    o3->set_w( 340 );

    m_glui_tf_path = m_glui_tffile->add_edittext( "File Path" );
    m_glui_tffile->add_column( false );
    /*
    GLUI_Button* b1 = m_glui_tffile->add_button("Browse...", 0);
    */
    m_glui_tf_path->set_w( 500 );
//    b1->set_w(150);

    GLUI_Button* b2 = m_glui_tffctl->add_button( "Export File", 0, cb_export_file );
    // Add Close Button 2016.2.14
    GLUI_Button* b4 = m_glui_tffctl->add_button( "Close", 1, cbTFEpanel_sw );

    m_glui_tffctl->add_column( false );
    GLUI_Button* b3 = m_glui_tffctl->add_button( "Import File", 0, cb_import_file );
    b2->set_w( 150 );
    b3->set_w( 150 );
    b4->set_w( 150 );

    this->m_glui_tf_name->set_int_val( m_current_tf );
    this->display_transferFunction( m_current_tf );

    show_tfe = true;
    return id;
}

void TFE_main::hide( void )
{
    ::ptfe = NULL;

    glutSetWindow( m_id );
    int x = glutGet( ( GLenum )GLUT_WINDOW_X );
    int y = glutGet( ( GLenum )GLUT_WINDOW_Y );
#if defined(WIN32)
    SuperClass::setPosition( x, y );
#elif defined(__APPLE__)
    SuperClass::setPosition( x, y );
#else
    SuperClass::setPosition( x, y - 28 );
#endif
    SuperClass::hide();
    show_tfe = false;
}

bool TFE_main::get_show_tfe( void )
{
    return show_tfe;
}

kvs::ScreenBase* TFE_main::screen( void )
{
    return ( m_screen );
}

const kvs::glut::ColorMapPalette* TFE_main::colorMapPalette( void ) const
{
    return ( m_color_map_palette );
}

const kvs::glut::OpacityMapPalette* TFE_main::opacityMapPalette( void ) const
{
    return ( m_opacity_map_palette );
}

const kvs::ColorMap TFE_main::colorMap( void ) const
{
    return ( m_color_map_palette->colorMap() );
}

const kvs::OpacityMap TFE_main::opacityMap( void ) const
{
    return ( m_opacity_map_palette->opacityMap() );
}

const kvs::TransferFunction TFE_main::transferFunction( void ) const
{
    kvs::TransferFunction transfer_function( this->colorMap(), this->opacityMap() );
    transfer_function.setRange( m_min_value, m_max_value );

    return ( transfer_function );
}

void TFE_main::setVolumeObject( const kvs::VolumeObjectBase* object )
{
    m_color_histogram->create( object );
    m_opacity_histogram->create( object );
}

kvs::visclient::ParamExTransFunc& TFE_main::doc( void )
{
    return m_doc;
}

const kvs::visclient::ParamExTransFunc& TFE_main::doc( void ) const
{
    return m_doc;
}

void TFE_main::exportFile( const std::string& fname, bool addition )
{
    std::ofstream ofs;
    if ( addition )
    {
        ofs.open( fname.c_str(), std::ofstream::app );
    }
    else
    {
        ofs.open( fname.c_str(), std::ofstream::out );
    }
    if ( ofs.fail() )
    {
        std::cerr << "Error: open file " << fname << std::endl;
        return;
    }

    ofs << "TF_RESOLUTION=" << m_doc.ext_transferFunctionResolution << std::endl;
    ofs << "TF_SYNTH=" << m_doc.ext_transferFunctionSynthesis << std::endl;
    for ( size_t n = 0; n < m_doc.transferFunction.size(); n++ )
    {
        std::stringstream ss;
        ss << "TF_NAME" << n + 1 << "_";

        const std::string tag_base = ss.str();
        ofs << tag_base << "VAR_C=" << m_doc.transferFunction[n].ColorVar << std::endl;
        ofs << tag_base << "MIN_C=" << m_doc.transferFunction[n].ColorVarMin << std::endl;
        ofs << tag_base << "MAX_C=" << m_doc.transferFunction[n].ColorVarMax << std::endl;
        ofs << tag_base << "VAR_O=" << m_doc.transferFunction[n].OpacityVar << std::endl;
        ofs << tag_base << "MIN_O=" << m_doc.transferFunction[n].OpacityVarMin << std::endl;
        ofs << tag_base << "MAX_O=" << m_doc.transferFunction[n].OpacityVarMax << std::endl;

        kvs::ColorMap::Table color_table = m_doc.transferFunction[n].colorMap().table();
        kvs::OpacityMap::Table opacity_table = m_doc.transferFunction[n].opacityMap().table();

        ofs << tag_base << "TABLE_C=";
        for ( size_t i = 0; i < color_table.size(); i++ )
        {
            ofs << static_cast<int>( color_table.at( i ) ) << ",";
        }
        ofs << std::endl;

        ofs << tag_base << "TABLE_O=";
        for ( size_t i = 0; i < opacity_table.size(); i++ )
        {
            ofs << opacity_table.at( i ) << ",";
        }
        ofs << std::endl;
    }

    ofs.close();

    std::cerr << "TransferFunction parameters are exported to " << fname << std::endl;
}

void TFE_main::importFile( const std::string& fname )
{
    ParamExTransFunc doc;
    ParamInfo param;
    bool stat;

    stat = param.LoadIN( fname );

    if ( !stat ) return;

    doc.ext_transferFunctionResolution = param.getInt( "tf_resolution" );
    doc.ext_transferFunctionSynthesis = param.getString( "tf_synthesis" );

    for ( size_t n = 0; n < doc.transferFunction.size(); n++ )
    {
        const size_t resolution = doc.ext_transferFunctionResolution;
        std::stringstream ss;
        ss << "TF_NAME" << n + 1 << "_";

        const std::string tag_base = ss.str();
        doc.transferFunction[n].setResolution( resolution );
        doc.transferFunction[n].ColorVar      = param.getString( tag_base + "VAR_C" );
        doc.transferFunction[n].ColorVarMin   = param.getFloat( tag_base + "MIN_C" );
        doc.transferFunction[n].ColorVarMax   = param.getFloat( tag_base + "MAX_C" );
        doc.transferFunction[n].OpacityVar    = param.getString( tag_base + "VAR_O" );
        doc.transferFunction[n].OpacityVarMin = param.getFloat( tag_base + "MIN_O" );
        doc.transferFunction[n].OpacityVarMax = param.getFloat( tag_base + "MAX_O" );

        std::string s_color = param.getString( tag_base + "TABLE_C" );
        std::string s_opacity = param.getString( tag_base + "TABLE_O" );

        std::replace( s_color.begin(), s_color.end(), ',', ' ' );
        std::replace( s_opacity.begin(), s_opacity.end(), ',', ' ' );

        std::stringstream ss_color( s_color );
        std::stringstream ss_opacity( s_opacity );

        kvs::ColorMap::Table color_table( resolution * 3 );
        kvs::OpacityMap::Table opacity_table( resolution );

        for ( size_t i = 0; i < resolution; i++ )
        {
            for ( size_t c = 0; c < 3; c++ )
            {
                int color_e;
                ss_color >> color_e;
                color_table.at( i * 3 + c ) = color_e;
            }
        }

        for ( size_t i = 0; i < resolution; i++ )
        {
            float opacity;
            ss_opacity >> opacity;
            opacity_table.at( i ) = opacity;
        }

        kvs::ColorMap color_map( color_table );
        kvs::OpacityMap opacity_map( opacity_table );
        kvs::TransferFunction tf( color_map, opacity_map );
        kvs::TransferFunction& tt = doc.transferFunction[n];
        tt = tf;
    }

    m_doc_initial = doc;
    m_doc = doc;
    m_command->isImportTransferFunctionParam = true;

    display_transferFunction( m_current_tf );
    this->redraw();

    std::cerr << "TransferFunction parameters are imported from " << fname << std::endl;
}

void TFE_main::importFileWithInit( const std::string& fname )
{
    this->importFile( fname );
    this->m_command->param.paramExTransFunc = m_doc;
}

void TFE_main::resolution( void )
{
    size_t resolution = m_glui_resolution->get_int_val();
    m_doc.ext_transferFunctionResolution = resolution;

    for ( size_t n = 0; n < m_doc.transferFunction.size(); n++ )
    {
        m_doc.transferFunction[n].setResolution( resolution );
    }

    this->display_transferFunction( m_current_tf );
    this->redraw();
}

void TFE_main::function_name( void )
{
    m_current_tf = m_glui_tf_name->get_int_val();
    this->display_transferFunction( m_current_tf );
}

void TFE_main::function_variable( void )
{
    const size_t n = m_current_tf;
    m_doc.transferFunction[n].ColorVar = m_glui_tf_var_c->get_text();
    m_doc.transferFunction[n].OpacityVar = m_glui_tf_var_o->get_text();
    m_doc.transferFunction[n].ColorVarMin = m_glui_tf_min_c->get_float_val();
    m_doc.transferFunction[n].ColorVarMax = m_glui_tf_max_c->get_float_val();
    m_doc.transferFunction[n].OpacityVarMin = m_glui_tf_min_o->get_float_val();
    m_doc.transferFunction[n].OpacityVarMax = m_glui_tf_max_o->get_float_val();
    m_doc.ext_transferFunctionSynthesis = m_glui_tf_synth->get_text();
    m_command->isImportTransferFunctionParam = true;
}

void TFE_main::reset( void )
{
    size_t resolution = m_doc.ext_transferFunctionResolution;
    for ( size_t n = 0; n < m_doc.transferFunction.size(); n++ )
    {
        kvs::TransferFunction* ptf = &m_doc.transferFunction[n];
        *ptf = kvs::TransferFunction( resolution );
    }
//  m_doc = m_doc_initial;

    this->display_transferFunction( m_current_tf );
    this->redraw();
}

void TFE_main::apply( void )
{
    this->m_command->param.paramExTransFunc = m_doc;
}

void TFE_main::import_file( void )
{
    std::string file_name = m_glui_tf_path->get_text();
    this->importFile( file_name );
}

void TFE_main::export_file( void )
{
    std::string file_name = m_glui_tf_path->get_text();
    this->exportFile( file_name, false );
}

void TFE_main::show_tfe_c1( void )
{
    m_tfe_c1->show();
    m_tfe_c1->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_c1->setTransferFunction( m_color_map_palette->colorMap() );
    m_tfe_c1->redraw();
}

void TFE_main::show_tfe_c2( void )
{
    m_tfe_c2->show();
    m_tfe_c2->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_c2->setEquation( "3*x-1.5", "1.4*sin(3.14*x)", "-3*x+1.5" );
    m_tfe_c2->redraw();
}

void TFE_main::show_tfe_c3( void )
{
    m_tfe_c3->show();
    m_tfe_c3->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_c3->setControlPointValue( 0, 0.00,  0,  0, 1.0 );
    m_tfe_c3->setControlPointValue( 1, 0.25,  0, 1.0, 1.0 );
    m_tfe_c3->setControlPointValue( 2, 0.50,  0, 1.0,  0 );
    m_tfe_c3->setControlPointValue( 3, 0.75, 1.0, 1.0,  0 );
    m_tfe_c3->setControlPointValue( 4, 1.00, 1.0,  0,  0 );
    m_tfe_c3->update();
    m_tfe_c3->redraw();
}

void TFE_main::show_tfe_c4( void )
{
    m_tfe_c4->show();
    m_tfe_c4->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_c4->setColorMapName( "Rainbow" );
    m_tfe_c4->redraw();
}

void TFE_main::show_tfe_o1( void )
{
    m_tfe_o1->show();
    m_tfe_o1->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_o1->setTransferFunction( m_opacity_map_palette->opacityMap() );
    m_tfe_o1->redraw();
}

void TFE_main::show_tfe_o2( void )
{
    m_tfe_o2->show();
    m_tfe_o2->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_o2->setEquation( "x" );
    m_tfe_o2->redraw();
}

void TFE_main::show_tfe_o3( void )
{
    m_tfe_o3->show();
    m_tfe_o3->setResolution( m_doc.ext_transferFunctionResolution );
    m_tfe_o3->setControlPointValue( 0, 0.00, 0.0 );
    m_tfe_o3->setControlPointValue( 1, 1.00, 1.0 );
    m_tfe_o3->update();
    m_tfe_o3->redraw();
}

void TFE_main::save_tfe_c1( void )
{
    const kvs::ColorMap cmap = m_tfe_c1->colorMap();
    this->m_color_map_palette->setColorMap( cmap );
    this->m_doc.transferFunction[m_current_tf].setColorMap( cmap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_c2( void )
{
    const kvs::ColorMap cmap = m_tfe_c2->colorMap();
    this->m_color_map_palette->setColorMap( cmap );
    this->m_doc.transferFunction[m_current_tf].setColorMap( cmap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_c3( void )
{
    const kvs::ColorMap cmap = m_tfe_c3->colorMap();
    this->m_color_map_palette->setColorMap( cmap );
    this->m_doc.transferFunction[m_current_tf].setColorMap( cmap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_c4( void )
{
    const kvs::ColorMap cmap = m_tfe_c4->colorMap();
    this->m_color_map_palette->setColorMap( cmap );
    this->m_doc.transferFunction[m_current_tf].setColorMap( cmap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_o1( void )
{
    const kvs::OpacityMap omap = m_tfe_o1->opacityMap();
    this->m_opacity_map_palette->setOpacityMap( omap );
    this->m_doc.transferFunction[m_current_tf].setOpacityMap( omap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_o2( void )
{
    const kvs::OpacityMap omap = m_tfe_o2->opacityMap();
    this->m_opacity_map_palette->setOpacityMap( omap );
    this->m_doc.transferFunction[m_current_tf].setOpacityMap( omap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::save_tfe_o3( void )
{
    const kvs::OpacityMap omap = m_tfe_o3->opacityMap();
    this->m_opacity_map_palette->setOpacityMap( omap );
    this->m_doc.transferFunction[m_current_tf].setOpacityMap( omap );
//  this->redraw();
    this->hide();
    this->show();
}

void TFE_main::display_transferFunction( const int n )
{
    kvs::ColorMap color_map = m_doc.transferFunction[n].colorMap();
    kvs::OpacityMap opacity_map = m_doc.transferFunction[n].opacityMap();

    m_color_map_palette->setColorMap( color_map );
    m_opacity_map_palette->setOpacityMap( opacity_map );

    if ( m_command->result.c_bins.size() > n )
    {
        m_color_histogram->setTable( m_command->result.c_bins[n] );
    }
    if ( m_command->result.o_bins.size() > n )
    {
        m_opacity_histogram->setTable( m_command->result.o_bins[n] );
    }

    m_glui_tf_var_c->set_text( m_doc.transferFunction[n].ColorVar );
    m_glui_tf_var_o->set_text( m_doc.transferFunction[n].OpacityVar );
    m_glui_tf_min_c->set_float_val( m_doc.transferFunction[n].ColorVarMin );
    m_glui_tf_max_c->set_float_val( m_doc.transferFunction[n].ColorVarMax );
    m_glui_tf_min_o->set_float_val( m_doc.transferFunction[n].OpacityVarMin );
    m_glui_tf_max_o->set_float_val( m_doc.transferFunction[n].OpacityVarMax );

    m_glui_resolution->set_int_val( m_doc.ext_transferFunctionResolution );
    m_glui_tf_synth->set_text( m_doc.ext_transferFunctionSynthesis );

    this->redraw();
    m_color_histogram->Redraw();
    m_opacity_histogram->Redraw();
}

void TFE_main::update_range_view( void )
{
    this->update_range_view( m_current_tf );
}

void TFE_main::update_range_view( const int n )
{
    kvs::ColorMap color_map = m_doc.transferFunction[n].colorMap();
    kvs::OpacityMap opacity_map = m_doc.transferFunction[n].opacityMap();
    const std::string nm = m_doc.transferFunction[n].Name;
    const std::string tag_c = nm + "_var_c";
    const std::string tag_o = nm + "_var_o";

    m_color_map_palette->setColorMap( color_map );
    m_opacity_map_palette->setOpacityMap( opacity_map );

    if ( m_command->result.c_bins.size() > n )
    {
        m_color_histogram->setTable( m_command->result.c_bins[n] );
    }
    if ( m_command->result.o_bins.size() > n )
    {
        m_opacity_histogram->setTable( m_command->result.o_bins[n] );
    }

    std::stringstream ss_min_c, ss_max_c;
    std::stringstream ss_min_o, ss_max_o;
    ss_min_c << "Server side range min: " << m_command->result.varRange.min( tag_c );
    ss_max_c << "Server side range max: " << m_command->result.varRange.max( tag_c );
    ss_min_o << "Server side range min: " << m_command->result.varRange.min( tag_o );
    ss_max_o << "Server side range max: " << m_command->result.varRange.max( tag_o );

    m_glui_rng_min_c->set_text( ss_min_c.str().c_str() );
    m_glui_rng_max_c->set_text( ss_max_c.str().c_str() );
    m_glui_rng_min_o->set_text( ss_min_o.str().c_str() );
    m_glui_rng_max_o->set_text( ss_max_o.str().c_str() );

    this->redraw();
    m_color_histogram->Redraw();
    m_opacity_histogram->Redraw();
}

void TFE_main::update_range_edit( void )
{
    this->update_range_edit( m_current_tf );
}

void TFE_main::update_range_edit( const int n )
{
    const std::string nm = m_doc.transferFunction[n].Name;
    const std::string tag_c = nm + "_var_c";
    const std::string tag_o = nm + "_var_o";

    const float var_min_c = m_command->result.varRange.min( tag_c );
    const float var_max_c = m_command->result.varRange.max( tag_c );
    const float var_min_o = m_command->result.varRange.min( tag_o );
    const float var_max_o = m_command->result.varRange.max( tag_o );

    m_glui_tf_min_c->set_float_val( var_min_c );
    m_glui_tf_max_c->set_float_val( var_max_c );
    m_glui_tf_min_o->set_float_val( var_min_o );
    m_glui_tf_max_o->set_float_val( var_max_o );

    this->redraw();
}

} // end of namespace visclient

} // end of namespace kvs
