/****************************************************************************/
/**
 *  @file GSubvolumeObjectUnstructuredVolume.cpp
 */
/****************************************************************************/
#include "GSubvolumeObjectUnstructuredVolume.h"
#include <Core/FileFormat/KVSML/UnstructuredVolumeObjectTag.h>
#include <Core/FileFormat/KVSML/NodeTag.h>
#include <Core/FileFormat/KVSML/CellTag.h>
#include <Core/FileFormat/KVSML/ValueTag.h>
#include <Core/FileFormat/KVSML/CoordTag.h>
#include <Core/FileFormat/KVSML/ConnectionTag.h>
#include "DataArrayTag.h"
#include <kvs/File>
#include <kvs/AnyValueArray>
#include <kvs/Type>
#include "Types.h"
#include <kvs/IgnoreUnusedVariable>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

namespace
{

/*===========================================================================*/
/**
 *  @brief  Returns the number of nodes per element.
 *  @param  cell_type [in] cell type as string
 *  @return number of nodes per element
 */
/*===========================================================================*/
inline const size_t GetNumberOfNodesPerElement( const std::string& cell_type )
{
    size_t nnodes = 0;
    if ( cell_type == "tetrahedra" ) nnodes = 4;
    else if ( cell_type == "quadratic tetrahedra" ) nnodes = 10;
    else if ( cell_type == "hexahedra" ) nnodes = 8;
    else if ( cell_type == "quadratic hexahedra" ) nnodes = 20;
    else if ( cell_type == "prism" ) nnodes = 6;
    else if ( cell_type == "pyramid" ) nnodes = 5;
    else if ( cell_type == "point" ) nnodes = 1;
    else if ( cell_type == "triangle" ) nnodes = 3;
    else if ( cell_type == "triangle2" ) nnodes = 6;
    else if ( cell_type == "quadratic" ) nnodes = 4;
    else if ( cell_type == "quadratic2" ) nnodes = 8;

    return ( nnodes );
}

} // end of namespace


namespace kvs
{

/*===========================================================================*/
/**
 *  @brief  Constructs a new GSubvolume object unstructured volume object class.
 */
/*===========================================================================*/
GSubvolumeObjectUnstructuredVolume::GSubvolumeObjectUnstructuredVolume( void )
{
}

/*===========================================================================*/
/**
 *  @brief  Constructs a new GSubvolume object unstructured volume object class by reading the given file.
 *  @param  filename [in] filename
 */
/*===========================================================================*/
GSubvolumeObjectUnstructuredVolume::GSubvolumeObjectUnstructuredVolume( const std::string& filename,
                                                                        int st, int vl )
{
    FILE* pfi_fp = NULL;
    char* pfi_name = NULL;
    int32_t cellType;
    int32_t numFiles;
    int32_t numIngredients;
    int32_t numSubVolumes;
    int32_t numNodes;
    int32_t numElements;
    int32_t startTimestep;
    int32_t elemNodes;
    size_t  offsetSubvol;
    size_t  offsetTimestep;
    size_t  offsetHalf;
    size_t  numTotalNodes;

    //-------------------------------------------------------------------------------------------------
    // load pfi file
    //-------------------------------------------------------------------------------------------------
    pfi_name = ( char* )malloc( filename.size() + 5 );
    sprintf( pfi_name, "%s.pfi", filename.c_str() );
    if ( ( pfi_fp = fopen( pfi_name, "rb" ) ) == NULL )
    {
        m_is_success = false;
        kvsMessageError( "Cannot read file '%s'.", pfi_name );
        free( pfi_name );
        return;
    }
    // Cell type
    fseek( pfi_fp, 4 * 2, SEEK_SET );
    fread( &cellType, 4, 1, pfi_fp );
    endian2::ltoh( cellType );
    switch ( cellType )
    {
    case 2:
        elemNodes = 3;
        setCellType( "triangle" );
        break;
    case 3:
        elemNodes = 4;
        setCellType( "quadratic" );
        break;
    case 4:
        elemNodes = 4;
        setCellType( "tetrahedra" );
        break;
    case 5:
        elemNodes = 5;
        setCellType( "pyramid" );
        break;
    case 6:
        elemNodes = 6;
        setCellType( "prism" );
        break;
    case 7:
        elemNodes = 8;
        setCellType( "hexahedra" );
        break;
    case 9:
        elemNodes = 6;
        setCellType( "triangle2" );
        break;
    case 10:
        elemNodes = 8;
        setCellType( "quadratic2" );
        break;
    case 11:
        elemNodes = 10;
        setCellType( "quadratic tetrahedra" );
        break;
    case 14:
        elemNodes = 20;
        setCellType( "quadratic hexahedra" );
        break;
    default:
        kvsMessageError( "Unknown cell type '%d'.", cellType );
        break;
    }
    // number of files
    fseek( pfi_fp, 4 * 4, SEEK_SET );
    fread( &numFiles, 4, 1, pfi_fp );
    endian2::ltoh( numFiles );
    // number of ingredients
    fseek( pfi_fp, 4 * 5, SEEK_SET );
    fread( &numIngredients, 4, 1, pfi_fp );
    endian2::ltoh( numIngredients );
    // start of Timestep
    fseek( pfi_fp, 4 * 6, SEEK_SET );
    fread( &startTimestep, 4, 1, pfi_fp );
    endian2::ltoh( startTimestep );
    // number of subvolumes
    fseek( pfi_fp, 4 * 8, SEEK_SET );
    fread( &numSubVolumes, 4, 1, pfi_fp );
    endian2::ltoh( numSubVolumes );
    // number of nodes
    fseek( pfi_fp, 4 * ( 15 + vl ), SEEK_SET );
    fread( &numNodes, 4, 1, pfi_fp );
    endian2::ltoh( numNodes );
    // number of elements
    fseek( pfi_fp, 4 * ( 15 + numSubVolumes + vl ), SEEK_SET );
    fread( &numElements, 4, 1, pfi_fp );
    endian2::ltoh( numElements );

    //-------------------------------------------------------------------------------------------------
    // search vol file
    //-------------------------------------------------------------------------------------------------
    FILE* fp = NULL;
    char* fname = NULL;
    int32_t flen = filename.size() + 21;
    kvs::ValueArray<kvs::Int32> range( ( size_t )2 );

    fname = ( char* )malloc( flen );
    if ( numFiles != numSubVolumes )
    {
        for ( int i = 1; i <= numFiles; i++ )
        {
            sprintf( fname, "%s_%07d_%07d.dat", filename.c_str(), i, numFiles );
            if ( fp = fopen( fname, "rb" ) )
            {
                fread( range.pointer(), 4, 2, fp );
#if __PBVR_BYTE_ORDER == __PBVR_BIG_ENDIAN
                range.swapByte();
#endif
                if ( range.at( 0 ) <= vl && range.at( 1 ) >= vl )
                {
                    // estimate file offset
                    int32_t nnode, nelem;
                    offsetSubvol   = 0;
                    offsetHalf     = 2;
                    offsetTimestep = 0;
                    numTotalNodes  = 0;
                    for ( int j = range.at( 0 ); j <= range.at( 1 ); j++ )
                    {
                        if ( j == vl )
                        {
                            offsetSubvol = offsetHalf;
                            offsetTimestep = numTotalNodes;
                        }
                        // number of nodes
                        fseek( pfi_fp, 4 * ( 15 + j ), SEEK_SET );
                        fread( &nnode, 4, 1, pfi_fp );
                        endian2::ltoh( nnode );
                        // number of elements
                        fseek( pfi_fp, 4 * ( 15 + numSubVolumes + j ), SEEK_SET );
                        fread( &nelem, 4, 1, pfi_fp );
                        endian2::ltoh( nelem );
                        offsetHalf += elemNodes * nelem + nnode * 3;
                        numTotalNodes += nnode * numIngredients;
                    }
                    offsetSubvol *= 4;
                    offsetTimestep += offsetHalf + ( st - startTimestep ) * numTotalNodes;
                    offsetTimestep *= 4;
                    break;
                }
                fclose( fp );
                fp = NULL;
            }
            else
            {
                fprintf( stderr, "File not exist. (%s)\n", fname );
            }
        }
    }
    else
    {
        sprintf( fname, "%s_%07d_%07d.dat", filename.c_str(), vl + 1, numFiles );
        if ( fp = fopen( fname, "rb" ) )
        {
            fread( range.pointer(), 4, 2, fp );
#if __PBVR_BYTE_ORDER == __PBVR_BIG_ENDIAN
            range.swapByte();
#endif
            if ( range.at( 0 ) <= vl && range.at( 1 ) >= vl )
            {
                // estimate file offset
                int32_t nnode, nelem;
                offsetSubvol   = 0;
                offsetHalf     = 2;
                offsetTimestep = 0;
                numTotalNodes  = 0;
                for ( int j = range.at( 0 ); j <= range.at( 1 ); j++ )
                {
                    if ( j == vl )
                    {
                        offsetSubvol = offsetHalf;
                        offsetTimestep = numTotalNodes;
                    }
                    // number of nodes
                    fseek( pfi_fp, 4 * ( 15 + j ), SEEK_SET );
                    fread( &nnode, 4, 1, pfi_fp );
                    endian2::ltoh( nnode );
                    // number of elements
                    fseek( pfi_fp, 4 * ( 15 + numSubVolumes + j ), SEEK_SET );
                    fread( &nelem, 4, 1, pfi_fp );
                    endian2::ltoh( nelem );
                    offsetHalf += elemNodes * nelem + nnode * 3;
                    numTotalNodes += nnode * numIngredients;
                }
                offsetSubvol *= 4;
                offsetTimestep += offsetHalf + ( st - startTimestep ) * numTotalNodes;
                offsetTimestep *= 4;
            }
            else
            {
                fclose( fp );
                fp = NULL;
            }
        }
        else
        {
            fprintf( stderr, "File not exist. (%s)\n", fname );
        }
    }

    //-------------------------------------------------------------------------------------------------
    // read vol file
    //-------------------------------------------------------------------------------------------------
    if ( fp )
    {
        // printf("input filename (%s): offset %d %d\n", fname, offsetSubvol, offsetTimestep);
        // printf(" timestep = %d: subvolume = %d: range = [%d, %d]\n", st, vl, range.at(0), range.at(1));
        // printf(" cellType=%d: numFiles=%d: numIngredients=%d: numSubvolumes=%d: numNodes=%d: numElements=%d\n",
        //        cellType, numFiles, numIngredients, numSubVolumes, numNodes, numElements);
        // connecion data
        if ( !m_connections.allocate( elemNodes * numElements ) )
        {
            kvsMessageError( "Cannot allocate memory for the connection data." );
            goto GSubvolumeObjectUnstructuredVolume_ERROR;
        }
        fseek( fp, offsetSubvol, SEEK_SET );
        if ( fread( m_connections.pointer(), 4, elemNodes * numElements, fp ) != elemNodes * numElements )
        {
            kvsMessageError( "File read error: connection data." );
        }
#if __PBVR_BYTE_ORDER == __PBVR_BIG_ENDIAN
        m_connections.swapByte();
#endif
        // coordinate data
        if ( !m_coords.allocate( numNodes * 3 ) )
        {
            kvsMessageError( "Cannot allocate memory for the coordinate data." );
            goto GSubvolumeObjectUnstructuredVolume_ERROR;
        }
        if ( fread( m_coords.pointer(), 4, numNodes * 3, fp ) != numNodes * 3 )
        {
            kvsMessageError( "File read error: coordinate data." );
        }
#if __PBVR_BYTE_ORDER == __PBVR_BIG_ENDIAN
        m_coords.swapByte();
#endif
        // value data
        if ( !m_values.allocate<kvs::Real32>( numNodes * numIngredients ) )
        {
            kvsMessageError( "Cannot allocate memory for the value data." );
            goto GSubvolumeObjectUnstructuredVolume_ERROR;
        }
        fseek( fp, offsetTimestep, SEEK_SET );
        if ( fread( m_values.pointer(), 4, numNodes * numIngredients, fp ) != numNodes * numIngredients )
        {
            kvsMessageError( "File read error: value data." );
        }
#if __PBVR_BYTE_ORDER == __PBVR_BIG_ENDIAN
        m_values.swapByte();
#endif
        // num nodes
        m_nnodes = numNodes;
        // veclen
        m_veclen = numIngredients;
        // num cells
        m_ncells = numElements;

    }
    else
    {
        kvsMessageError( "Cannot open file '%s'.", fname );
        goto GSubvolumeObjectUnstructuredVolume_ERROR;
    }

    m_is_success = true;
    if ( pfi_fp ) fclose( pfi_fp );
    if ( pfi_name ) free( pfi_name );
    if ( fp ) fclose( fp );
    if ( fname ) free( fname );
    return;

GSubvolumeObjectUnstructuredVolume_ERROR:
    if ( pfi_fp ) fclose( pfi_fp );
    if ( pfi_name ) free( pfi_name );
    if ( fp ) fclose( fp );
    if ( fname ) free( fname );
    m_is_success = false;
}

/*===========================================================================*/
/**
 *  @brief  Destructs the GSubvolume object unstructured volume object class.
 */
/*===========================================================================*/
GSubvolumeObjectUnstructuredVolume::~GSubvolumeObjectUnstructuredVolume( void )
{
}

/*===========================================================================*/
/**
 *  @brief  Returns the cell type as string.
 *  @return cell type
 */
/*===========================================================================*/
const std::string GSubvolumeObjectUnstructuredVolume::cellType( void ) const
{
    return ( m_cell_type );
}

/*===========================================================================*/
/**
 *  @brief  Returns the vector length.
 *  @return vector length
 */
/*===========================================================================*/
const size_t GSubvolumeObjectUnstructuredVolume::veclen( void ) const
{
    return ( m_veclen );
}

/*===========================================================================*/
/**
 *  @brief  Returns the number of nodes.
 *  @return number of nodes
 */
/*===========================================================================*/
const size_t GSubvolumeObjectUnstructuredVolume::nnodes( void ) const
{
    return ( m_nnodes );
}

/*===========================================================================*/
/**
 *  @brief  Returns the number of cells.
 *  @return number of cells
 */
/*===========================================================================*/
const size_t GSubvolumeObjectUnstructuredVolume::ncells( void ) const
{
    return ( m_ncells );
}

/*===========================================================================*/
/**
 *  @brief  Returns the value array.
 *  @return value array
 */
/*===========================================================================*/
const kvs::AnyValueArray& GSubvolumeObjectUnstructuredVolume::values( void ) const
{
    return ( m_values );
}

/*===========================================================================*/
/**
 *  @brief  Returns the coordinate array.
 *  @return coordinate array
 */
/*===========================================================================*/
const kvs::ValueArray<kvs::Real32>& GSubvolumeObjectUnstructuredVolume::coords( void ) const
{
    return ( m_coords );
}

/*===========================================================================*/
/**
 *  @brief  Returns the connection array.
 *  @return connection array
 */
/*===========================================================================*/
const kvs::ValueArray<kvs::UInt32>& GSubvolumeObjectUnstructuredVolume::connections( void ) const
{
    return ( m_connections );
}

/*===========================================================================*/
/**
 *  @brief  Sets a cell type.
 *  @param  cell_type [in] cell type
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setCellType( const std::string cell_type )
{
    m_cell_type = cell_type;
}

/*===========================================================================*/
/**
 *  @brief  Sets a vector length.
 *  @param  veclen [in] vector length
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setVeclen( const size_t veclen )
{
    m_veclen = veclen;
}

/*===========================================================================*/
/**
 *  @brief  Sets a number of nodes.
 *  @param  nnodes [in] number of nodes
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setNNodes( const size_t nnodes )
{
    m_nnodes = nnodes;
}

/*===========================================================================*/
/**
 *  @brief  Sets a number of cells.
 *  @param  ncells [in] number of cells
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setNCells( const size_t ncells )
{
    m_ncells = ncells;
}

/*===========================================================================*/
/**
 *  @brief  Sets a value array.
 *  @param  values [in] value array
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setValues( const kvs::AnyValueArray& values )
{
    m_values = values;
}

/*===========================================================================*/
/**
 *  @brief  Sets a coordinate array.
 *  @param  coords [in] coordinate array
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setCoords( const kvs::ValueArray<kvs::Real32>& coords )
{
    m_coords = coords;
}

/*===========================================================================*/
/**
 *  @brief  Sets a connection array.
 *  @param  connections [in] connection array
 */
/*===========================================================================*/
void GSubvolumeObjectUnstructuredVolume::setConnections( const kvs::ValueArray<kvs::UInt32>& connections )
{
    m_connections = connections;
}

const bool GSubvolumeObjectUnstructuredVolume::read( const std::string& filename )
{
    return false;
}
const bool GSubvolumeObjectUnstructuredVolume::write( const std::string& filename )
{
    return false;
}
} // end of namespace kvs
