
#include "TransferFunctionSynthesizer.h"
#include <fstream>
#include <limits>

using FuncParser::Variable;
using FuncParser::Variables;
using FuncParser::Function;
using FuncParser::FunctionParser;

static const std::string str_coordX( "X" );
static const std::string str_coordY( "Y" );
static const std::string str_coordZ( "Z" );

TransferFunctionSynthesizer::TransferFunctionSynthesizer()
    : m_synth_func_str( "t1" )
{
    m_ptf_map_c["t1"] = pbvr::TransferFunction();
    m_ptf_map_o["t1"] = pbvr::TransferFunction();
    m_tfq_map_c["t1"] = "q1";
    m_tfq_map_o["t1"] = "q1";
    m_qmap["q1"]    = 0;

    this->create();
}

TransferFunctionSynthesizer::TransferFunctionSynthesizer( const TransferFunctionSynthesizer& s )
{
    copy( s );
}

TransferFunctionSynthesizer::TransferFunctionSynthesizer( const TransferFunctionSynthesizer* s )
{
    if ( s )
    {
        copy( *s );
    }
    else
    {
        TransferFunctionSynthesizer();
    }
}

TransferFunctionSynthesizer::~TransferFunctionSynthesizer()
{
}

const TransferFunctionMap& TransferFunctionSynthesizer::color_transfunc_map() const
{
    return m_ptf_map_c;
}

const TransferFunctionMap& TransferFunctionSynthesizer::opacity_transfunc_map() const
{
    return m_ptf_map_o;
}

const TransfuncQuantityBindMap& TransferFunctionSynthesizer::color_transfunc_bind_map() const
{
    return m_tfq_map_c;
}

const TransfuncQuantityBindMap& TransferFunctionSynthesizer::opacity_transfunc_bind_map() const
{
    return m_tfq_map_o;
}

const FunctionStringMap& TransferFunctionSynthesizer::volfuncstr_map() const
{
    return m_volfuncstr_map;
}

const QuantityMap& TransferFunctionSynthesizer::quantity_map() const
{
    return m_qmap;
}

const FunctionMap& TransferFunctionSynthesizer::func_map() const
{
    return m_volfunc_map;
}

const std::string& TransferFunctionSynthesizer::synth_function_string() const
{
    return m_synth_func_str;
}

void TransferFunctionSynthesizer::setMaxOpacity( const float opacity )
{
    m_max_opacity = opacity;
}

void TransferFunctionSynthesizer::setMaxDensity( const float density )
{
    m_max_density = density;
}

void TransferFunctionSynthesizer::setSamplingVolumeInverse( const float sampling_volume_inverse )
{
    m_sampling_volume_inverse = sampling_volume_inverse;
}

float TransferFunctionSynthesizer::getMaxOpacity() const
{
    return m_max_opacity;
}

float TransferFunctionSynthesizer::getMaxDensity() const
{
    return m_max_density;
}

float TransferFunctionSynthesizer::getSamplingVolumeInverse() const
{
    return m_sampling_volume_inverse;
}

VariableRange& TransferFunctionSynthesizer::variableRange()
{
    return m_variable_range;
}

const VariableRange& TransferFunctionSynthesizer::variableRange() const
{
    return m_variable_range;
}

void TransferFunctionSynthesizer::mergeVarRange( const TransferFunctionSynthesizer& tfs )
{
    m_variable_range.merge( tfs.m_variable_range );
}

void TransferFunctionSynthesizer::copy( const TransferFunctionSynthesizer& s )
{
    this->setColorTransferFunctionMap( s.color_transfunc_map() );
    this->setOpacityTransferFunctionMap( s.opacity_transfunc_map() );
    this->setColorTransfuncQuantityBindMap( s.color_transfunc_bind_map() );
    this->setOpacityTransfuncQuantityBindMap( s.opacity_transfunc_bind_map() );
    this->setVolumeFunctionStringMap( s.volfuncstr_map() );
    this->setQuantityMap( s.quantity_map() );
    this->setSynthFunctionString( s.synth_function_string() );
    this->create();

    this->setMaxOpacity( s.getMaxOpacity() );
    this->setMaxDensity( s.getMaxDensity() );
    this->setSamplingVolumeInverse( s.getSamplingVolumeInverse() );

    m_variable_range = s.m_variable_range;
}

TransferFunctionSynthesizer& TransferFunctionSynthesizer::operator= ( const TransferFunctionSynthesizer& s )
{
    copy( s );
    return *this;
}

void TransferFunctionSynthesizer::setValue( const std::string& nm, const float value )
{
    VariableMap::const_iterator i = m_var_map.find( nm );
    if ( i != m_var_map.end() )
    {
        m_var_map[nm] = value;
    }
}

float TransferFunctionSynthesizer::getValue( const std::string& nm )
{
    VariableMap::const_iterator i = m_var_map.find( nm );
    float ret = std::numeric_limits<float>::quiet_NaN();
    if ( i != m_var_map.end() )
    {
        ret = i->second.eval();
    }
    return ret;
}

void TransferFunctionSynthesizer::setTransferFunctionMap( const TransferFunctionMap& ptf_map )
{
    setColorTransferFunctionMap( ptf_map );
    setOpacityTransferFunctionMap( ptf_map );
}

void TransferFunctionSynthesizer::setColorTransferFunctionMap( const TransferFunctionMap& ptf_map )
{
    for ( TransferFunctionMap::const_iterator i = ptf_map.begin(); i != ptf_map.end(); i++ )
    {
        m_ptf_map_c[i->first] = pbvr::TransferFunction( i->second );
    }
}

void TransferFunctionSynthesizer::setOpacityTransferFunctionMap( const TransferFunctionMap& ptf_map )
{
    for ( TransferFunctionMap::const_iterator i = ptf_map.begin(); i != ptf_map.end(); i++ )
    {
        m_ptf_map_o[i->first] = pbvr::TransferFunction( i->second );
    }
}

void TransferFunctionSynthesizer::setTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map )
{
    setColorTransfuncQuantityBindMap( tfq_map );
    setOpacityTransfuncQuantityBindMap( tfq_map );
}

void TransferFunctionSynthesizer::setColorTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map )
{
    for ( TransfuncQuantityBindMap::const_iterator i = tfq_map.begin(); i != tfq_map.end(); i++ )
    {
        m_tfq_map_c[i->first] = i->second;
    }
}

void TransferFunctionSynthesizer::setOpacityTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map )
{
    for ( TransfuncQuantityBindMap::const_iterator i = tfq_map.begin(); i != tfq_map.end(); i++ )
    {
        m_tfq_map_o[i->first] = i->second;
    }
}

void TransferFunctionSynthesizer::setVolumeFunctionStringMap( const FunctionStringMap& volfuncstr_map )
{
    const std::string sp( " \t" );
    for ( FunctionStringMap::const_iterator i = volfuncstr_map.begin(); i != volfuncstr_map.end(); i++ )
    {
        std::string fs( i->second );
        size_t s = std::string::npos;
        for ( std::string::const_iterator c = sp.begin(); c != sp.end(); c++ )
        {
            while ( ( s = fs.find( *c, 0 ) ) != std::string::npos )
            {
                fs.replace( s, 1, "" );
            }
        }
        m_volfuncstr_map[i->first] = fs;
    }
}

void TransferFunctionSynthesizer::setQuantityMap( const QuantityMap& qmap )
{
    for ( QuantityMap::const_iterator i = qmap.begin(); i != qmap.end(); i++ )
    {
        m_qmap[i->first] = i->second;
    }
}

void TransferFunctionSynthesizer::setSynthFunctionString( const std::string& synth_func_str )
{
    m_synth_func_str = synth_func_str;
}

void TransferFunctionSynthesizer::create()
{
    // clear containers
    m_vars.clear();
    m_var_map.clear();
    m_volfunc_map.clear();

    m_variable_range.clear();

    // Variables for X,Y,Z coord
    addVariant( str_coordX );
    addVariant( str_coordY );
    addVariant( str_coordZ );

    // Vairables for quantities
    for ( QuantityMap::iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        addVariant( nm );
    }

    // variables which have evaluation value of primitive transfer function
    for ( TransferFunctionMap::iterator i = m_ptf_map_c.begin(); i != m_ptf_map_c.end(); i++ )
    {
        const std::string& nm = i->first;
        addVariant( nm );
    }

    // initialize and parse volume shape functions
    initVolumeFunction();
    registVolumeFunction();

    // initialize synthesized transfer function
    FunctionParser synth_func_parse( m_synth_func_str, m_synth_func_str.size() + 1 );
    FunctionParser::Error err = synth_func_parse.express( m_synth_func, m_vars );
}

void TransferFunctionSynthesizer::optimize()
{
    TransfuncQuantityBindMap new_tfq_c, new_tfq_o;
    FunctionStringMap new_volfuncstr_map;

    // volume functions
    for ( TransfuncQuantityBindMap::iterator b = m_tfq_map_c.begin(); b != m_tfq_map_c.end(); b++ )
    {
        const std::string& fn = b->first;
        const std::string& qn = b->second;
        const std::string& qs = m_volfuncstr_map[qn];
        for ( FunctionStringMap::iterator f = m_volfuncstr_map.begin(); f != m_volfuncstr_map.end(); f++ )
        {
            const std::string& fs = f->second;
            if ( fs == qs )
            {
                new_tfq_c[fn] = f->first;
                break;
            }
        }
    }

    for ( TransfuncQuantityBindMap::iterator b = m_tfq_map_o.begin(); b != m_tfq_map_o.end(); b++ )
    {
        const std::string& fn = b->first;
        const std::string& qn = b->second;
        const std::string& qs = m_volfuncstr_map[qn];
        for ( FunctionStringMap::iterator f = m_volfuncstr_map.begin(); f != m_volfuncstr_map.end(); f++ )
        {
            const std::string& fs = f->second;
            if ( fs == qs )
            {
                new_tfq_o[fn] = f->first;
                break;
            }
        }
    }

    for ( FunctionStringMap::iterator f = m_volfuncstr_map.begin(); f != m_volfuncstr_map.end(); f++ )
    {
        const std::string& fn = f->first;
        const std::string& fs = f->second;
        for ( TransfuncQuantityBindMap::iterator b = new_tfq_c.begin(); b != new_tfq_c.end(); b++ )
        {
            if ( b->second == fn )
            {
                new_volfuncstr_map[fn] = fs;
                break;
            }
        }
        for ( TransfuncQuantityBindMap::iterator b = new_tfq_o.begin(); b != new_tfq_o.end(); b++ )
        {
            if ( b->second == fn )
            {
                new_volfuncstr_map[fn] = fs;
                break;
            }
        }
    }

    m_tfq_map_c.clear();
    m_tfq_map_o.clear();
    m_volfuncstr_map.clear();

    setColorTransfuncQuantityBindMap( new_tfq_c );
    setOpacityTransfuncQuantityBindMap( new_tfq_o );
    setVolumeFunctionStringMap( new_volfuncstr_map );
}

void TransferFunctionSynthesizer::optimize2()
{
    std::vector<std::string> v;
    const std::string& fs = m_synth_func.str();
    for ( TransferFunctionMap::iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
    {
        const std::string& fn = t->first;
        if ( fs.find( fn , 0 ) == std::string::npos )
        {
            v.push_back( fn );
        }
    }

    for ( size_t i = 0; i < v.size(); i++ )
    {
        const std::string& fn = v[i];
        m_ptf_map_c.erase( fn );
        m_ptf_map_o.erase( fn );
        m_tfq_map_c.erase( fn );
        m_tfq_map_o.erase( fn );
    }
}

void TransferFunctionSynthesizer::clear()
{
    m_vars.clear();
    m_var_map.clear();
    m_volfunc_map.clear();

    m_variable_range.clear();

    m_ptf_map_c.clear();
    m_ptf_map_o.clear();
    m_tfq_map_c.clear();
    m_tfq_map_o.clear();
    m_volfuncstr_map.clear();
    m_qmap.clear();

    m_ts_vec.clear();
    m_synth_func_str.clear();
}

//kawamura
void TransferFunctionSynthesizer::multivariateMinMaxValues(
    pbvr::CellBase<kvs::Real32>* cell, const size_t cell_index, const size_t nnodes )
{
    //セル頂点について一つづつ最大最小値を計算.
    for ( size_t i = 0; i < nnodes; i++ )
    {
        // qは変量 q = q(X,Y,Z)
        // fは変量の合成式 f = f(X,Y,Z,q1,..,qn)
        // tは１次元伝達関数 t = t(f).

        // 頂点の X,Y,Zをセット
        m_var_map[str_coordX] = cell->vertices()[i].x();
        m_var_map[str_coordY] = cell->vertices()[i].y();
        m_var_map[str_coordZ] = cell->vertices()[i].z();

        // 頂点のq1,q2,..,qnをセット.
        for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
        {
            const std::string& qname = q->first;//qの変量名.
            const size_t cid = q->second;//変量qのコンポーネントID
            cell->bindCell( cell_index, cid );//コンポーネントIDを指定してセルをバインド.
            m_var_map[qname] = cell->scalars()[i];//コンポーネントIDに対応する頂点の変量qをセット.
        }

        // 頂点のX,Y,Z,q1,q2,..qnからfを計算.
        for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
        {
            const std::string& fname = i->first;//fの変量名.
            const Function& f = i->second;//変量f
            //どこかでexpressコマンドを走らせているならevalコマンドでfの値が得られる
            m_var_map[fname] = f.eval();
        }

        //不透明度関数に対応したfの値を最大最小値計算メソッドに放り込む
        for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
        {
            const std::string& tname = t->first;//１次元伝達関数tの名前
            const std::string& fname = m_tfq_map_o.find( tname )->second;//１次元伝達関数tの引数fの名前
            const float f_value = m_var_map[fname].eval();//fの値
            const std::string tag = tname + "_var_o";
            m_variable_range.setValue( tag, f_value );
        }

        //色関数に対応したfの値を最大最小値計算メソッドに放り込む
        for ( TransferFunctionMap::const_iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
        {
            const std::string& tname = t->first;//１次元伝達関数tの名前
            const std::string& fname = m_tfq_map_c.find( tname )->second;//１次元伝達関数tの引数fの名前
            const float f_value = m_var_map[fname].eval();//fの値
            const std::string tag = tname + "_var_c";
            m_variable_range.setValue( tag, f_value );
        }
    }
}

//kawamura
void TransferFunctionSynthesizer::multivariateMinMaxAveragedValues(
    pbvr::CellBase<kvs::Real32>* cell, const size_t cell_index, const size_t nnodes )
{
    //セル頂点について一つづつ最大最小値を計算.
    // qは変量 q = q(X,Y,Z)
    // fは変量の合成式 f = f(X,Y,Z,q1,..,qn)
    // tは１次元伝達関数 t = t(f).

    // 頂点の X,Y,Zをセット
    m_var_map[str_coordX] = cell->vertices()[0].x();
    m_var_map[str_coordY] = cell->vertices()[0].y();
    m_var_map[str_coordZ] = cell->vertices()[0].z();

    //不透明度関数に対応したfの値を最大最小値計算メソッドに放り込む
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
    {
        const std::string& tname = t->first;//１次元伝達関数tの名前
        const std::string& fname = m_tfq_map_o.find( tname )->second;//１次元伝達関数tの引数fの名前
        const float f_value = m_var_map[fname].eval();//fの値
        const std::string tag = tname + "_var_o";
        m_variable_range.setValue( tag, f_value );
    }

    //色関数に対応したfの値を最大最小値計算メソッドに放り込む
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
    {
        const std::string& tname = t->first;//１次元伝達関数tの名前
        const std::string& fname = m_tfq_map_c.find( tname )->second;//１次元伝達関数tの引数fの名前
        const float f_value = m_var_map[fname].eval();//fの値
        const std::string tag = tname + "_var_c";
        m_variable_range.setValue( tag, f_value );
    }
}


float TransferFunctionSynthesizer::CalculateOpacity(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    // qは変量 q = q(X,Y,Z)
    // fは変量の合成式 f = f(X,Y,Z,q1,..,qn)
    // tは１次元伝達関数 t = t(f).

    // set X,Y,Z
    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    // X,Y,Zからq1,q2,..,qnを計算.
    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->scalar();
    }

    // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
    {
        const std::string& nm = t->first;//１次元伝達関数tの名前
        const pbvr::TransferFunction& f = t->second;//１次元伝達関数t

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_o.find( nm )->second;//１次元伝達関数tの引数fの名前
        const float scalar = m_volfunc_map[qnm].eval();//fの値

        // set value to stab variable
        if ( scalar < f.minValue() )
        {
            m_var_map[nm] = f.opacityMap().at( f.minValue() );
        }
        else if ( f.maxValue() < scalar )
        {
            m_var_map[nm] = f.opacityMap().at( f.maxValue() );
        }
        else
        {
            m_var_map[nm] = f.opacityMap().at( scalar );
        }
    }

    // eval synthesized transfer function
    float opacity = m_synth_func.eval();
    if ( opacity > 1.0 ) opacity = 1.0;
    if ( opacity < 0.0 ) opacity = 0.0;

    return opacity;
}


float TransferFunctionSynthesizer::CalculateAveragedOpacityBatch(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->averagedScalar();
    }

    // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
    {
        const std::string& nm = t->first;
        const pbvr::TransferFunction& f = t->second;

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_o.find( nm )->second;
        const float scalar = m_volfunc_map[qnm].eval();

        // set value to stab variable
        if ( scalar < f.minValue() )
        {
            m_var_map[nm] = f.opacityMap().at( f.minValue() );
        }
        else if ( f.maxValue() < scalar )
        {
            m_var_map[nm] = f.opacityMap().at( f.maxValue() );
        }
        else
        {
            m_var_map[nm] = f.opacityMap().at( scalar );
        }
    }

    // eval synthesized transfer function
    float opacity = m_synth_func.eval();
    if ( opacity > 1.0 ) opacity = 1.0;
    if ( opacity < 0.0 ) opacity = 0.0;

    return opacity;
}

float TransferFunctionSynthesizer::CalculateAveragedOpacity(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->averagedScalar();
    }

#if 0
    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        m_var_map[nm] = f.eval();
    }
#endif

    // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
    {
        const std::string& nm = t->first;
        const pbvr::TransferFunction& f = t->second;

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_o.find( nm )->second;
        const float scalar = m_volfunc_map[qnm].eval();
        const std::string tag = nm + "_var_o";
        m_variable_range.setValue( tag, scalar );
        m_var_map[tag] = scalar;

        // set value to stab variable
        if ( scalar < f.minValue() )
        {
            m_var_map[nm] = f.opacityMap().at( f.minValue() );
        }
        else if ( f.maxValue() < scalar )
        {
            m_var_map[nm] = f.opacityMap().at( f.maxValue() );
        }
        else
        {
            m_var_map[nm] = f.opacityMap().at( scalar );
        }
    }

    // eval synthesized transfer function
    float opacity = m_synth_func.eval();
    if ( opacity > 1.0 ) opacity = 1.0;
    if ( opacity < 0.0 ) opacity = 0.0;

    return opacity;
}

float TransferFunctionSynthesizer::CalculateGravityPointOpacity(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    // qは変量 q = q(X,Y,Z)
    // fは変量の合成式 f = f(X,Y,Z,q1,..,qn)
    // tは１次元伝達関数 t = t(f).

    // set X,Y,Z
    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    // X,Y,Zからq1,q2,..,qnを計算.
    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->localGravityPointValue();
    }

    // X,Y,Z,q1,q2,..qnからfを計算.
    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        m_var_map[nm] = f.eval();
    }

    // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
    for ( TransferFunctionMap::const_iterator t = m_ptf_map_o.begin(); t != m_ptf_map_o.end(); t++ )
    {
        const std::string& nm = t->first;//１次元伝達関数tの名前
        const pbvr::TransferFunction& f = t->second;//１次元伝達関数t

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_o.find( nm )->second;//１次元伝達関数tの引数fの名前
        const float scalar = m_var_map[qnm].eval();//fの値
        const std::string tag = nm + "_var_o";
        m_variable_range.setValue( tag, scalar );
        m_var_map[tag] = scalar;

        // set value to stab variable
        if ( scalar < f.minValue() )
        {
            m_var_map[nm] = f.opacityMap().at( f.minValue() );
        }
        else if ( f.maxValue() < scalar )
        {
            m_var_map[nm] = f.opacityMap().at( f.maxValue() );
        }
        else
        {
            m_var_map[nm] = f.opacityMap().at( scalar );
        }
    }

    // eval synthesized transfer function
    float opacity = m_synth_func.eval();
    if ( opacity > 1.0 ) opacity = 1.0;
    if ( opacity < 0.0 ) opacity = 0.0;

    return opacity;
}

kvs::RGBColor TransferFunctionSynthesizer::CalculateAveragedColor(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    kvs::UInt8 col[3];
    str_color_map c_map;

    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->averagedScalar();
    }

#if 0
    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        m_var_map[nm] = f.eval();
    }
#endif

    for ( TransferFunctionMap::const_iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
    {
        const std::string& nm = t->first;
        const pbvr::TransferFunction& f = t->second;

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_c.find( nm )->second;
//      const float scalar = m_var_map[qnm].eval();
        const float scalar = m_volfunc_map[qnm].eval();
        const std::string tag = nm + "_var_c";
        m_variable_range.setValue( tag, scalar );
        m_var_map[tag] = scalar;

        // reference transfer function
        c_map[nm] = ( scalar < f.minValue() ) ? f.colorMap().at( f.minValue() ) :
                    ( scalar > f.maxValue() ) ? f.colorMap().at( f.maxValue() ) :
                    f.colorMap().at( scalar );
    }

    //kawamura
    float color[3];
    // eval synthesized transfer function
    for ( unsigned c = 0; c < 3; c++ ) // for each RGB element
    {
        for ( str_color_map::iterator i = c_map.begin(); i != c_map.end(); i++ )
        {
            const std::string& nm = i->first;
            const kvs::RGBColor& cv = i->second;

            float r = ( float )cv.r() / 255.0;
            float g = ( float )cv.g() / 255.0;
            float b = ( float )cv.b() / 255.0;

            m_var_map[nm] = ( c == 0 ) ? r :
                            ( c == 1 ) ? g :
                            b;
            /*m_var_map[nm] = (c == 0)? cv.r():
                            (c == 1)? cv.g():
                            cv.b();*/

        }

        color[c] = m_synth_func.eval();
        col[c] = ( kvs::UInt8 )( color[c] * 255.0 );
        //col[c] = m_synth_func.eval();
    }

    return kvs::RGBColor( col );
}

kvs::RGBColor TransferFunctionSynthesizer::CalculateColor(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    kvs::UInt8 col[3];
    str_color_map c_map;

    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->scalar();
    }

#if 0
    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        m_var_map[nm] = f.eval();
    }
#endif

    for ( TransferFunctionMap::const_iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
    {
        const std::string& nm = t->first;
        const pbvr::TransferFunction& f = t->second;

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_c.find( nm )->second;
//      const float scalar = m_var_map[qnm].eval();
        const float scalar = m_volfunc_map[qnm].eval();

        // reference transfer function
        c_map[nm] = ( scalar < f.minValue() ) ? f.colorMap().at( f.minValue() ) :
                    ( scalar > f.maxValue() ) ? f.colorMap().at( f.maxValue() ) :
                    f.colorMap().at( scalar );
    }

    //kawamura
    float color[3];
    // eval synthesized transfer function
    for ( unsigned c = 0; c < 3; c++ ) // for each RGB element
    {
        for ( str_color_map::iterator i = c_map.begin(); i != c_map.end(); i++ )
        {
            const std::string& nm = i->first;
            const kvs::RGBColor& cv = i->second;

            float r = ( float )cv.r() / 255.0;
            float g = ( float )cv.g() / 255.0;
            float b = ( float )cv.b() / 255.0;

            m_var_map[nm] = ( c == 0 ) ? r :
                            ( c == 1 ) ? g :
                            b;
            /*m_var_map[nm] = (c == 0)? cv.r():
                            (c == 1)? cv.g():
                            cv.b();*/

        }

        color[c] = m_synth_func.eval();
        col[c] = ( kvs::UInt8 )( color[c] * 255.0 );
        //col[c] = m_synth_func.eval();
    }

    return kvs::RGBColor( col );
}

kvs::RGBColor TransferFunctionSynthesizer::CalculateGravityPointColor(
    pbvr::CellBase<kvs::Real32>* cell,
    const size_t n,
    const kvs::Vector3f& coord )
{
    kvs::UInt8 col[3];
    str_color_map c_map;

    m_var_map[str_coordX] = coord.x();
    m_var_map[str_coordY] = coord.y();
    m_var_map[str_coordZ] = coord.z();

    for ( QuantityMap::const_iterator q = m_qmap.begin(); q != m_qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        const size_t qi = q->second;
        cell->bindCell( n, qi );

        m_var_map[nm] = cell->localGravityPointValue();
    }

    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        m_var_map[nm] = f.eval();
    }

    for ( TransferFunctionMap::const_iterator t = m_ptf_map_c.begin(); t != m_ptf_map_c.end(); t++ )
    {
        const std::string& nm = t->first;
        const pbvr::TransferFunction& f = t->second;

        // physical quantity of the cell
        const std::string& qnm = m_tfq_map_c.find( nm )->second;
        const float scalar = m_var_map[qnm].eval();
        const std::string tag = nm + "_var_c";
        m_variable_range.setValue( tag, scalar );
        m_var_map[tag] = scalar;

        // reference transfer function
        c_map[nm] = ( scalar < f.minValue() ) ? f.colorMap().at( f.minValue() ) :
                    ( scalar > f.maxValue() ) ? f.colorMap().at( f.maxValue() ) :
                    f.colorMap().at( scalar );
    }

    //kawamura
    float color[3];
    // eval synthesized transfer function
    for ( unsigned c = 0; c < 3; c++ ) // for each RGB element
    {
        for ( str_color_map::iterator i = c_map.begin(); i != c_map.end(); i++ )
        {
            const std::string& nm = i->first;
            const kvs::RGBColor& cv = i->second;

            float r = ( float )cv.r() / 255.0;
            float g = ( float )cv.g() / 255.0;
            float b = ( float )cv.b() / 255.0;

            m_var_map[nm] = ( c == 0 ) ? r :
                            ( c == 1 ) ? g :
                            b;
            /*m_var_map[nm] = (c == 0)? cv.r():
                            (c == 1)? cv.g():
                            cv.b();*/

        }

        color[c] = m_synth_func.eval();
        col[c] = ( kvs::UInt8 )( color[c] * 255.0 );
        //col[c] = m_synth_func.eval();
    }

    return kvs::RGBColor( col );
}

void TransferFunctionSynthesizer::addVariant( const std::string& nm )
{
    m_var_map[nm] = Variable();
    Variable& var = m_var_map[nm];

    m_tcp_vec.push_back_strcpy( nm );

    var.tag( m_tcp_vec.back() );
    m_vars.push_back( var );
}

void TransferFunctionSynthesizer::initVolumeFunction()
{
    for ( FunctionStringMap::iterator i = m_volfuncstr_map.begin(); i != m_volfuncstr_map.end(); i++ )
    {
        const std::string& nm  = i->first;
        const std::string& str = i->second;
        m_volfunc_map[nm] = Function();
        FunctionParser parse( const_cast<std::string&>( str ), str.size() + 1 );
        FunctionParser::Error err = parse.express( m_volfunc_map[nm], m_vars );
        if ( err != FunctionParser::ERR_NONE )
        {
            std::cerr << FunctionParser::error_type_to_string[err] << std::endl;
        }
    }
}

void TransferFunctionSynthesizer::registVolumeFunction()
{
    for ( FunctionMap::const_iterator i = m_volfunc_map.begin(); i != m_volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        addVariant( nm );
    }
}

