#ifndef __TRANSFER_FUNCTION_SYNTHESIZER_H__
#define __TRANSFER_FUNCTION_SYNTHESIZER_H__

#include <map>
#include <string>

#include "TransferFunction.h"
#include <kvs/RGBColor>
#include "CellBase.h"

#include "function.h"
#include "function_op.h"
#include "function_parser.h"

#include "VariableRange.h"

// Dictionaries Type-definition
typedef std::map<std::string, pbvr::TransferFunction> TransferFunctionMap;
typedef std::map<std::string, FuncParser::Variable>  VariableMap;
typedef std::map<std::string, std::string>           FunctionStringMap;
typedef std::map<std::string, FuncParser::Function>  FunctionMap;
typedef std::map<std::string, size_t>                QuantityMap;
typedef std::map<std::string, std::string>           TransfuncQuantityBindMap;

typedef std::vector<std::string>                     TagStringVec;

class TagCharPtrVec : public std::vector<char*>
{
public:
    void push_back_strcpy( const std::string& nm )
    {
        char* tmp = new char[ nm.size() + 1 ];
        strcpy( tmp, nm.c_str() );
        this->push_back( tmp );
    }
    TagCharPtrVec() {};
    ~TagCharPtrVec()
    {
        for ( std::vector<char*>::iterator t = this->begin(); t != this->end(); t++ )
        {
            delete[] ( *t );
        }
        this->clear();
    }
};

class TransferFunctionSynthesizer
{

    typedef std::map<std::string, kvs::RGBColor> str_color_map;

protected:
    // [ キー, 値 ]
    // qは変量 q = q(X,Y,Z)
    // fは変量の合成式 f = f(X,Y,Z,q1,..,qn)
    // tは１次元伝達関数 t = t(f). tは色関数cと不透明度関数oの組.
    FuncParser::Variables m_vars;
    VariableMap           m_var_map;// [ 変量名(X,Y,Z,q1,..,qn,f1,..,fn), 変量 ].
    FunctionMap           m_volfunc_map;//[ 変量名(f1,..,fn), 変量の合成式 ]

    TransferFunctionMap      m_ptf_map_c; // [ 色関数名(c1,..,c5), 色関数 ]
    TransferFunctionMap      m_ptf_map_o; // [ 不透明度関数名(o1,..,o5), 不透明度関数 ]
    TransfuncQuantityBindMap m_tfq_map_c; // [ 色関数名(c1,..,c5), 色関数cの引数 ]
    TransfuncQuantityBindMap m_tfq_map_o; // [ 不透明度関数名(o1,..,o5), 不透明度関数oの引数 ]
    FunctionStringMap        m_volfuncstr_map;
    QuantityMap              m_qmap; // [ 変量名(q1,..,qn), コンポーネントID ]

    TagStringVec             m_ts_vec;
    TagCharPtrVec            m_tcp_vec;

    std::string          m_synth_func_str;
    FuncParser::Function m_synth_func;

    float m_max_opacity;
    float m_max_density;
    float m_sampling_volume_inverse;

    VariableRange m_variable_range;

public:

    TransferFunctionSynthesizer();
    explicit TransferFunctionSynthesizer( const TransferFunctionSynthesizer& s );
    TransferFunctionSynthesizer( const TransferFunctionSynthesizer* s );

    virtual ~TransferFunctionSynthesizer();

    const TransferFunctionMap& color_transfunc_map() const;
    const TransferFunctionMap& opacity_transfunc_map() const;

    const TransfuncQuantityBindMap& color_transfunc_bind_map() const;
    const TransfuncQuantityBindMap& opacity_transfunc_bind_map() const;

    const FunctionStringMap& volfuncstr_map() const;

    const QuantityMap& quantity_map() const;

    const FunctionMap& func_map() const;

    const std::string& synth_function_string() const;

    //kawamura
    void multivariateMinMaxValues( pbvr::CellBase<kvs::Real32>* cell,
                                   const size_t cell_index,
                                   const size_t nnodes );

    void multivariateMinMaxAveragedValues( pbvr::CellBase<kvs::Real32>* cell,
                                           const size_t cell_index,
                                           const size_t nnodes );

    void setMaxOpacity( const float opacity );
    void setMaxDensity( const float density );
    void setSamplingVolumeInverse( const float sampling_volume_inverse );

    float getMaxOpacity() const;
    float getMaxDensity() const;
    float getSamplingVolumeInverse() const;

    VariableRange& variableRange();
    const VariableRange& variableRange() const;
    void  mergeVarRange( const TransferFunctionSynthesizer& tfs );

    void copy( const TransferFunctionSynthesizer& s );
    TransferFunctionSynthesizer& operator= ( const TransferFunctionSynthesizer& s );

    void setValue( const std::string& nm, const float value );
    float getValue( const std::string& nm );

    void setTransferFunctionMap( const TransferFunctionMap& ptf_map );
    void setColorTransferFunctionMap( const TransferFunctionMap& ptf_map );
    void setOpacityTransferFunctionMap( const TransferFunctionMap& ptf_map );

    void setTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map );
    void setColorTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map );
    void setOpacityTransfuncQuantityBindMap( const TransfuncQuantityBindMap& tfq_map );

    void setVolumeFunctionStringMap( const FunctionStringMap& volfuncstr_map );
    void setQuantityMap( const QuantityMap& qmap );
    void setSynthFunctionString( const std::string& synth_func_str );

    void optimize();
    void optimize2();
    void create();
    void clear();

    float CalculateOpacity( pbvr::CellBase<kvs::Real32>* cell,
                            const size_t n,
                            const kvs::Vector3f& coord );
    float CalculateAveragedOpacity( pbvr::CellBase<kvs::Real32>* cell,
                                    const size_t n,
                                    const kvs::Vector3f& coord );
    float CalculateAveragedOpacityBatch( pbvr::CellBase<kvs::Real32>* cell,
                                         const size_t n,
                                         const kvs::Vector3f& coord );
    float CalculateGravityPointOpacity( pbvr::CellBase<kvs::Real32>* cell,
                                        const size_t n,
                                        const kvs::Vector3f& coord );

    kvs::RGBColor CalculateColor( pbvr::CellBase<kvs::Real32>* cell,
                                  const size_t n,
                                  const kvs::Vector3f& coord );
    kvs::RGBColor CalculateAveragedColor( pbvr::CellBase<kvs::Real32>* cell,
                                          const size_t n,
                                          const kvs::Vector3f& coord );
    kvs::RGBColor CalculateGravityPointColor( pbvr::CellBase<kvs::Real32>* cell,
                                              const size_t n,
                                              const kvs::Vector3f& coord );

private:
    void addVariant( const std::string& nm );
    void initVolumeFunction();
    void registVolumeFunction();
};

#endif // __TRANSFER_FUNCTION_SYNTHESIZER_H__

