/*****************************************************************************/
/**
 *  main.cpp
 *  CPU Server Program without CUDA, without KVS, without OpenGL.
 */
/*****************************************************************************/

#include "ParticleTransferServer.h"
#include "ParticleTransferProtocol.h"

#include "PointObject.h"
#include <kvs/CommandLine>
#include <kvs/Camera>
#include "TransferFunction.h"
#include <kvs/Matrix33>
#include <kvs/RotationMatrix33>

#include "common.h"

#include "PointObjectGenerator.h"

#include "Argument.h"

#include <kvs/AVSField>
#include <kvs/Timer>
#include "KVSMLObjectPointWriter.h"
//#include "KVSMLObjectPointMPIWriter.h"
#include "JobDispatcher.h"
#ifndef CPU_VER
#include "JobCollector.h"
#endif
#include "FilterInfo.h"
#include "ParamInfo.h"
#ifndef CPU_VER
#include "mpi.h"
#endif
#ifdef KMATH
#include <kmath_random.h>
#endif

#include <cassert>
#include <signal.h> /* 140319 for client stop by Ctrl+c */
#include <sys/stat.h>
#include <kvs/File>

#include "ExtendedTransferFunction.h"
#include "TransferFunctionSynthesizerCreator.h"
#include "VariableRange.h"
#include "RangeEstimater.h"

#include "timer_simple.h"

#include <kvs/Compiler>
#ifdef KVS_COMPILER_VC
#include <direct.h>
#define mkdir( dir, mode ) _mkdir( dir )
#endif

#include "UnstructuredVolumeImporter.h"
#include "CellByCellParticleGenerator.h"

bool useAllNodes = true;
#ifdef KMATH
KMATH_Random km_random;
#endif

inline const size_t GetRevisedSubpixelLevel(
    const size_t subpixel_level,
    const size_t repetition_level )
{
    return ( static_cast<size_t>( subpixel_level * std::sqrt( ( double )repetition_level ) + 0.5f ) );
}

class PointObjectCreator
{
private:

#if 0 //TEST_DELETE
    kvs::AVSField* m_field;
#endif

    pbvr::UnstructuredVolumeObject* m_volume;

    pbvr::PointObjectGenerator m_generator;

    int m_mpi_rank;

    const FilterInfo* m_fi;

public:

    PointObjectCreator( void ) {}

    ~PointObjectCreator()
    {
#if 0 //TEST_DELETE
        if ( m_field != NULL )
        {
            /* 131022 removed for segmentation fault in batch mode
               m_field->close();
               delete m_field;
               delete m_volume;
            */
        }
#endif
    }

    void setFilterInfo( const FilterInfo& fi )
    {
        m_fi = &fi;
    }

    void setGT5D( void )
    {
#if 0 //TEST_DELETE
        this->read_GT5D_connections_and_values();
        m_volume->setCoords( m_field->coords() );
        m_volume->updateMinMaxCoords();
#endif
    }

    void setGT5D_full( int rank )
    {
#if 0 //TEST_DELETE
        m_mpi_rank = rank;
        this->read_GT5D_connections_and_values();

        const kvs::Vector3ui resol = m_field->dim();
        const kvs::UInt32 nnodes = resol.x() * resol.y() * resol.z();
        const kvs::Vector3ui elem = resol - kvs::Vector3ui( 1 );
        const kvs::UInt32 nelem = elem.x() * elem.y() * elem.z();

        kvs::ValueArray<kvs::Real32> coords( nnodes * 3 );
        kvs::Matrix33f mat( kvs::YRotationMatrix33<float>( ( float )rank * 60.0 ) );
        size_t i3 = 0;
        for ( size_t i = 0; i < nnodes; i++, i3 += 3 )
        {
            kvs::Vector3f vec;
            vec.x() = m_field->coords()[i3];
            vec.y() = m_field->coords()[i3 + 1];
            vec.z() = m_field->coords()[i3 + 2];

            vec = mat * vec;
            coords[i3] = vec.x();
            coords[i3 + 1] = vec.y();
            coords[i3 + 2] = vec.z();

        }

        m_volume->setCoords( coords );
        m_volume->updateMinMaxCoords();
        std::cout << rank << ":ucd vol\n" << *m_volume << std::endl;
#endif
    }

    void progressValues( void )
    {
#if 0 //TEST_DELETE
        m_field->progress();
        m_volume->setValues( kvs::AnyValueArray( m_field->values( 0 ) ) );
        m_volume->updateMinMaxValues();

#endif
    }

public:

    pbvr::PointObject* run( Argument& param, kvs::Camera& camera, int timeStep )
    {
        m_generator.setFinlterInfo( *m_fi );
        if ( param.pyramid )
        {
            m_generator.createPyramidPointObject
            ( param, camera, param.subpixel_level, param.sampling_step );
        }
        else if ( param.GT5D )
        {
            m_generator.createGT5D(
                param, camera, param.subpixel_level, param.sampling_step, m_volume );
        }
        else if ( param.HTTR )
        {
            m_generator.createHTTR(
                param, camera, param.subpixel_level, param.sampling_step, timeStep );
        }
        else if ( param.input_data == "" )
        {
            m_generator.createQuadraticHexahedra
            ( param, camera, param.subpixel_level, param.sampling_step );
        }
        else
        {
            //sleep(15);

            struct stat st;
            if ( stat( param.input_data.c_str(), &st ) )
            {
                std::cout << "Error. read failed:" << param.input_data << std::endl;
                exit( 1 );
            }
            m_generator.createFromFile( param, camera, param.subpixel_level, param.sampling_step );
        }
        return m_generator.getPointObject();
    }

    pbvr::PointObject* run( Argument& param, kvs::Camera& camera, int timeStep, int st, int vl )
    {
        m_generator.setFinlterInfo( *m_fi );
        m_generator.createFromFile( param, camera, param.subpixel_level, param.sampling_step, st, vl );
        return m_generator.getPointObject();
    }

protected:

    void read_GT5D_connections_and_values( void )
    {
#if 0 //TEST_DELETE
        std::string header = "/home/kawamura/Data/GT5D_large/610x610x128x100-2/pd3d.fld";
        std::string coord = "/home/kawamura/Data/GT5D_large/610x610x128x100-2/co3d.dat";
        std::string variable = "/home/kawamura/Data/GT5D_large/610x610x128x100-2/pd3d.dat";

        m_field = new kvs::AVSField( header, coord, variable );
        m_field->progress();

        m_volume = new pbvr::UnstructuredVolumeObject;

        const kvs::Vector3ui resol = m_field->dim();
        const kvs::UInt32 nnodes = resol.x() * resol.y() * resol.z();
        const kvs::Vector3ui elem = resol - kvs::Vector3ui( 1 );
        const kvs::UInt32 nelem = elem.x() * elem.y() * elem.z();

        m_volume->setVeclen( 1 );
        m_volume->setNNodes( nnodes );
        m_volume->setNCells( nelem );
        m_volume->setCellType( pbvr::UnstructuredVolumeObject::Hexahedra );
        m_volume->setValues( kvs::AnyValueArray( m_field->values( 0 ) ) );

        const kvs::UInt32 line_size = resol.x();
        const kvs::UInt32 slice_size = line_size * resol.y();
        kvs::ValueArray<kvs::UInt32> connections( nelem * 8 );

        kvs::UInt32 m_index[8];
        kvs::UInt32* connec = connections.pointer();
        std::cout << m_mpi_rank << ":start of gen. conne." << std::endl;
        for ( size_t k = 0; k < elem.z(); k++ )
        {
            for ( size_t j = 0; j < elem.y(); j++ )
            {
                for ( size_t i = 0; i < elem.x(); i++ )
                {
                    m_index[0] = i + j * line_size + k * slice_size;
                    m_index[1] = m_index[0] + 1;
                    m_index[2] = m_index[1] + line_size;
                    m_index[3] = m_index[0] + line_size;
                    m_index[4] = m_index[0] + slice_size;
                    m_index[5] = m_index[1] + slice_size;
                    m_index[6] = m_index[2] + slice_size;
                    m_index[7] = m_index[3] + slice_size;

                    for ( size_t p = 0; p < 8; p++ )
                    {
                        *connec = m_index[p];
                        connec++;
                    }
                }
            }
        }
        m_volume->setConnections( connections );
        m_volume->updateMinMaxValues();
#endif
    }
};

/* 140319 for client stop by Ctrl+c */
bool SigServer = false;
void sig_handler( int sig )
{
    SigServer = true;
}
/* 140319 for client stop by Ctrl+c */

bool isDirectory( std::string directory_path )
{
#if defined ( WIN32 )
    WIN32_FIND_DATAA find_data;
    HANDLE hFind = FindFirstFileA( directory_path.c_str(), &find_data );
    if ( hFind == INVALID_HANDLE_VALUE )
    {
        return ( false );
    }
    return ( ( find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 );
#else
    struct stat filestat;
    if ( stat( directory_path.c_str(), &filestat ) )
    {
        return ( false );
    }
    return ( filestat.st_mode & S_IFDIR );
#endif
}

inline pbvr::UnstructuredVolumeObject* CreateVolumeData( const Argument& param,
                                                        const FilterInfo& fi,
                                                        int& steps, const int& subvols )
{
    if ( fi.fileType == 1 || fi.fileType == 2 )
    {
        pbvr::UnstructuredVolumeObject* volume = new pbvr::UnstructuredVolumeImporter( param.input_data_base,
                                                                                     fi.fileType, steps, subvols );
        volume->setMinMaxValues( fi.minValue, fi.maxValue );
        volume->setMinMaxObjectCoords( fi.minObjectCoord, fi.maxObjectCoord );
        volume->setMinMaxExternalCoords( fi.minObjectCoord, fi.maxObjectCoord );

        return volume;
    }
    else
    {
        std::stringstream suffix;
        suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << ( steps )
               << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( subvols + 1 )
               << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;

        //std::string input_data = param.input_data_base + suffix.str() + ".kvsml";
        kvs::File ifpx( fi.filePath );
        std::string input_data = ifpx.pathName() + ifpx.Separator()
                                 + ifpx.baseName() + suffix.str() + ".kvsml";
        pbvr::UnstructuredVolumeObject* volume = new pbvr::UnstructuredVolumeImporter( input_data );

        volume->setMinMaxValues( fi.minValue, fi.maxValue );
        volume->setMinMaxObjectCoords( fi.minObjectCoord, fi.maxObjectCoord );
        volume->setMinMaxExternalCoords( fi.minObjectCoord, fi.maxObjectCoord );

        return volume;
    }
}


inline float CalculateSamplingStep( const FilterInfoLst& fil )
{
    float max_coord_length = kvs::Math::Max<float>( fil.total_maxObjectCoord.x() - fil.total_minObjectCoord.x(),
                                                    fil.total_maxObjectCoord.y() - fil.total_minObjectCoord.y(),
                                                    fil.total_maxObjectCoord.z() - fil.total_minObjectCoord.z() );
    return 0.1 * max_coord_length;
}

//kawamura2: This calculates optimized subpixel level.
inline size_t CalculateSubpixelLevel( const Argument& param,
                                      const FilterInfoLst& fil,
                                      const kvs::Camera& camera )
{
    namespace Generator = pbvr::CellByCellParticleGenerator;
    pbvr::UnstructuredVolumeObject* volume;
    double total_volume = 0.0;
    float density_lev1;//kawamura2: particle density for subpixel_level=1
    int steps = fil.total_staSteps;
    int subvols = 0;

    //Total Volume Calculation
#ifndef CPU_VER
    int rank = MPI::COMM_WORLD.Get_rank();
    int nprocs = MPI::COMM_WORLD.Get_size();
#else
    int rank = 0;
    int nprocs = 1;
#endif

    if ( rank == 0 )
    {
        density_lev1 = 0.0f;
        for ( int n = 0; n < fil.m_list.size(); n++ )
        {
            const FilterInfo& fi = fil.m_list[n];
            TIMER_STA( 16 );
            volume = CreateVolumeData( param, fi, steps, subvols );
            TIMER_END( 16 );
            TIMER_STA( 17 );
            double local_volume = Generator::CalculateTotalVolume( volume );
            TIMER_END( 17 );
            TIMER_STA( 18 );
            //kawamura2
            density_lev1 += Generator::CalculateGreatDensity( &camera, volume, 1,
                                                              param.sampling_step ) * local_volume;
            TIMER_END( 18 );
            total_volume += local_volume;
            delete volume;
        } // end of for(n)
        density_lev1 /= total_volume;
    }
#ifndef CPU_VER
    TIMER_STA( 19 );
    MPI_Bcast( &density_lev1, 1, MPI_FLOAT, 0, MPI_COMM_WORLD ); //kawamura2
    TIMER_END( 19 );
#endif

    for ( subvols = subvols + 1; subvols < fil.total_numSubVolumes; subvols++ )
    {
        int xvl, fidx;
        fidx = fil.GetFileIdx( subvols, xvl );
        const FilterInfo& fi = fil.m_list[fidx];

        if ( subvols % nprocs == rank )
        {
            TIMER_STA( 16 );
            volume = CreateVolumeData( param, fi, steps, xvl );
            TIMER_END( 16 );
            TIMER_STA( 17 );
            total_volume += Generator::CalculateTotalVolume( volume );
            TIMER_END( 17 );

            delete volume;
        }
    }

#ifndef CPU_VER
    TIMER_STA( 19 );
    MPI_Allreduce( MPI_IN_PLACE, &total_volume, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
    TIMER_END( 19 );
#endif

    //kawamura2 : Calculation of optimized subpixel level
    float particle_limit = param.particle_limit;
    float nparticles_lev1 = total_volume * density_lev1;
    float subpixel_level = sqrt( particle_limit / nparticles_lev1 );
    if ( subpixel_level < 1 ) subpixel_level = 1;
    else if ( subpixel_level > 12 ) subpixel_level = 12;

    return static_cast<size_t>( subpixel_level + 0.5f );
}

inline float CalculateDensityFactor( const Argument& param,
                                     const FilterInfo& fi,
                                     const kvs::Camera& camera )
{
    namespace Generator = pbvr::CellByCellParticleGenerator;
    pbvr::UnstructuredVolumeObject* volume;
    double total_volume = 0.0;
    float great_density;
    int steps = fi.staSteps;
    int subvols = 0;
#ifndef CPU_VER
    int rank = MPI::COMM_WORLD.Get_rank();
    int nprocs = MPI::COMM_WORLD.Get_size();
#else
    int rank = 0;
    int nprocs = 1;
#endif

    if ( rank == 0 )
    {
        TIMER_STA( 16 );
        volume = CreateVolumeData( param, fi, steps, subvols );
        TIMER_END( 16 );
        TIMER_STA( 17 );
        total_volume += Generator::CalculateTotalVolume( volume );
        TIMER_END( 17 );
        TIMER_STA( 18 );
        great_density = Generator::CalculateGreatDensity( &camera, volume, param.subpixel_level,
                                                          param.sampling_step );
        TIMER_END( 18 );

        delete volume;
    }
#ifndef CPU_VER
    TIMER_STA( 19 );
    MPI_Bcast( &great_density, 1, MPI_FLOAT, 0, MPI_COMM_WORLD );
    TIMER_END( 19 );
#endif


    for ( subvols = subvols + 1; subvols < fi.numSubVolumes; subvols++ )
    {
        if ( subvols % nprocs == rank )
        {
            TIMER_STA( 16 );
            volume = CreateVolumeData( param, fi, steps, subvols );
            TIMER_END( 16 );
            TIMER_STA( 17 );
            total_volume += Generator::CalculateTotalVolume( volume );
            TIMER_END( 17 );

            delete volume;
        }
    }
#ifndef CPU_VER
    TIMER_STA( 19 );
    MPI_Allreduce( MPI_IN_PLACE, &total_volume, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
    TIMER_END( 19 );
#endif

    float total_nparticles = great_density * static_cast<float>( total_volume );
    float density_factor = static_cast<float>( param.particle_limit ) * 1000000 / total_nparticles;

    if ( density_factor > 1.0 ) density_factor = 1.0;

    return density_factor;
}

int main( int argc, char** argv )
{
#ifndef CPU_VER
    MPI::Init( argc, argv );
#endif
    TIMER_INIT();
    TIMER_STA( 1 );
#ifdef KMATH
#ifndef CPU_VER
    km_random.init( MPI_COMM_WORLD );
#else
    km_random.init();
#endif
    km_random.seed( 1 );
#endif
    Argument param( argc, argv );
    FilterInfoLst fil;
    TransferFunctionSynthesizerCreator transfunc_creator;

    kvs::Camera camera;
    kvs::Timer timer;
    int retval = 0;
    int mpi_rank = 0;
    std::vector<PointObjectCreator> point_creator_lst;
    //PointObjectCreator point_creator;
    pbvr::PointObject* object = NULL;
    std::string output, outdir;
    std::string pout = "PARTICLE_OUTDIR";
    std::string prfx = "PARTICLE_SERVER_PREFIX";


#ifndef CPU_VER
    int rank = MPI::COMM_WORLD.Get_rank();
#else
    int rank = 0;
#endif

    if ( param.batch == true )
    {

        TIMER_STA( 2 );
        TIMER_STA( 11 );
        if ( getenv( pout.c_str() ) )
        {
            outdir = std::string( getenv( pout.c_str() ) );
        }
        else
        {
            outdir = "";
        }

        kvs::File outdir_pfx( param.output_data_base );
#if defined ( WIN32 )
        outdir = outdir_pfx.pathName();
#else
        outdir = outdir_pfx.pathName() + outdir_pfx.Separator();
#endif
        if ( isDirectory( outdir ) )
        {
#if defined ( WIN32 )
            outdir = outdir + outdir_pfx.Separator();
#endif
            if ( outdir_pfx.fileName() != "" )
            {
                output = outdir + std::string( outdir_pfx.fileName().c_str() );
            }
            else
            {
                std::cerr << "Error \"-pout\" : " << "Prefix does not exist" << std::endl;
#ifdef KMATH
                km_random.finalize();
#endif
#ifndef CPU_VER
                MPI::Finalize();
#endif
                return 0;
            }
        }
        else
        {
            if ( rank == 0 ) std::cerr << outdir << " : " << "The directory does not exist" << std::endl;
#ifdef KMATH
            km_random.finalize();
#endif
#ifndef CPU_VER
            MPI::Finalize();
#endif
            return 0;
        }

        JobDispatcher jd;
#ifndef CPU_VER
        JobCollector  jc( &jd );
        if ( ! param.batch_fjoin )
            jc.SetBatch( true );
#endif

        TIMER_STA( 12 );
        bool pfi_Exist = false;
        std::string pflfile, pfifile;
        if ( param.input_data_base.substr( param.input_data_base.size() - 3 ) == "pfl" )
        {
            pflfile = param.input_data_base;
            kvs::File pfl( pflfile );
            if ( pfl.isExisted() )
            {
                fil.LoadPFL( pflfile );
            }
        }
        else
        {
            pfifile = param.input_data_base + ".pfi";
            kvs::File pfi( pfifile );
            pflfile = param.input_data_base + ".pfl";
            kvs::File pfl( pflfile );
            if ( pfl.isExisted() )
            {
                fil.LoadPFL( pflfile );
            }
            else if ( pfi.isExisted() )
            {
                fil.LoadPFL( pfifile );
            }
        }
        if ( param.batch_spec_ts )
            fil.CropTimestep( param.batch_ts[0], param.batch_ts[1] );

        if ( fil.m_list.size() > 0 )
        {
            point_creator_lst.clear();
            for ( int idx = 0; idx < fil.m_list.size(); idx++ )
            {
                PointObjectCreator point_creator;
                if ( param.GT5D == true ) point_creator.setGT5D();
                point_creator.setFilterInfo( fil.m_list[idx] );
                point_creator_lst.push_back( point_creator );
            }

            transfunc_creator.setFilterInfo( fil.m_list[0] );

            if ( rank == 0 )
            {
                std::cout << " start step = "         << fil.total_staSteps
                          << " end step = "           << fil.total_endSteps
                          << " time step = "          << fil.total_numSteps
                          << " subvolume division = " << fil.total_numSubVolumes
                          << std::endl;
            }

            if ( param.hasOption( "pout" ) )
            {
                std::string minmax_file = output + "_pfi_coords_minmax.txt";
                FILE* fp = fopen( minmax_file.c_str(), "w" );
                if ( fp )
                {
                    fprintf( fp, "%f %f %f %f %f %f\n",
                             fil.total_minObjectCoord[0],
                             fil.total_minObjectCoord[1], fil.total_minObjectCoord[2],
                             fil.total_maxObjectCoord[0],
                             fil.total_maxObjectCoord[1], fil.total_maxObjectCoord[2] );
                    fclose( fp );
                }
            }
            transfunc_creator.setRange( "t1", fil.total_minValue, fil.total_maxValue );
            pfi_Exist = true;
        }
        TIMER_END( 12 );


        if ( !param.hasOption( "L" ) ) param.latency_threshold = -1.0;
        int st, vl, wid = 0;

        TIMER_STA( 13 );
        if ( param.param_file != "" )
        {
            ParamInfo pa;
            if ( pa.LoadIN( param.param_file ) )
            {
                if ( !param.hasOption( "sl" ) && pa.getState( "SUB_PIXEL_LEVEL" ) )
                    param.subpixel_level = ( size_t )pa.getInt( "SUB_PIXEL_LEVEL" );
                if ( !param.hasOption( "rl" ) && pa.getState( "REPEAT_LEVEL" ) )
                    param.repeat_level = ( size_t )pa.getInt( "REPEAT_LEVEL" );
                if ( !param.hasOption( "plimit" ) && pa.getState( "PARTICLE_LIMIT" ) )
                    param.particle_limit = pa.getInt( "PARTICLE_LIMIT" );
                if ( pa.getState( "RESOLUTION_WIDTH" ) ) param.windowWidth = pa.getInt( "RESOLUTION_WIDTH" );
                if ( pa.getState( "RESOLUTION_HEIGHT" ) ) param.windowHeight = pa.getInt( "RESOLUTION_HEIGHT" );

                if ( !pfi_Exist )
                {
                    std::string input = pa.getString( "PFI_PATH_SERVER" );
                    size_t pos;

                    if ( ( pos = input.find( ".pfi" ) ) != std::string::npos )
                    {
                        param.input_data_base = input.substr( 0, pos );
                    }
                    else if ( ( pos = input.find( ".pfl" ) ) != std::string::npos )
                    {
                        param.input_data_base = input.substr( 0, pos );
                    }
                    else
                    {
                        param.input_data_base = input;
                    }

                    pfifile = param.input_data_base + ".pfi";
                    pflfile = param.input_data_base + ".pfl";
                    kvs::File pfi_pa( pfifile );
                    kvs::File pfl_pa( pflfile );
                    if ( pfl_pa.isExisted() )
                    {
                        fil.LoadPFL( pflfile );
                    }
                    else if ( pfi_pa.isExisted() )
                    {
                        fil.LoadPFL( pfifile );
                    }
                    if ( param.batch_spec_ts )
                        fil.CropTimestep( param.batch_ts[0], param.batch_ts[1] );

                    if ( fil.m_list.size() > 0 )
                    {
                        point_creator_lst.clear();
                        for ( int idx = 0; idx < fil.m_list.size(); idx++ )
                        {
                            PointObjectCreator point_creator;
                            if ( param.GT5D == true ) point_creator.setGT5D();
                            point_creator.setFilterInfo( fil.m_list[idx] );
                            point_creator_lst.push_back( point_creator );
                        }

                        transfunc_creator.setFilterInfo( fil.m_list[0] );

                        if ( rank == 0 )
                        {
                            std::cout << " time step = "          << fil.total_numSteps
                                      << " subvolume division = " << fil.total_numSubVolumes
                                      << std::endl;
                        }

                        if ( param.hasOption( "pout" ) )
                        {
                            std::string minmax_file = output + "_pfi_coords_minmax.txt";
                            FILE* fp = fopen( minmax_file.c_str(), "w" );
                            if ( fp )
                            {
                                fprintf( fp, "%f %f %f %f %f %f\n",
                                         fil.total_minObjectCoord[0],
                                         fil.total_minObjectCoord[1],
                                         fil.total_minObjectCoord[2],
                                         fil.total_maxObjectCoord[0],
                                         fil.total_maxObjectCoord[1],
                                         fil.total_maxObjectCoord[2] );
                                fclose( fp );
                            }
                        }
                        transfunc_creator.setRange( "t1", fil.total_minValue, fil.total_maxValue );
                        pfi_Exist = true;
                    }
                }

                if ( !pfi_Exist )
                {
                    if ( rank == 0 ) std::cerr << "Error: pfifile doesn't exist" << std::endl;
#ifdef KMATH
                    km_random.finalize();
#endif
#ifndef CPU_VER
                    MPI::Finalize();
#endif
                    return 0;
                }

                if ( !param.hasOption( "crop" ) && pa.getState( "CROP_TYPE" ) )
                {
                    switch ( pa.getInt( "CROP_TYPE" ) )
                    {
                    case 1:
                    {
                        float crop_tmp[6];
                        crop_tmp[0] = fil.total_minObjectCoord.x();
                        crop_tmp[1] = fil.total_minObjectCoord.y();
                        crop_tmp[2] = fil.total_minObjectCoord.z();
                        crop_tmp[3] = fil.total_maxObjectCoord.x();
                        crop_tmp[4] = fil.total_maxObjectCoord.y();
                        crop_tmp[5] = fil.total_maxObjectCoord.z();
                        if ( pa.getState( "CROP_CXMIN" ) ) crop_tmp[0] = pa.getFloat( "CROP_CXMIN" );
                        if ( pa.getState( "CROP_CYMIN" ) ) crop_tmp[1] = pa.getFloat( "CROP_CYMIN" );
                        if ( pa.getState( "CROP_CZMIN" ) ) crop_tmp[2] = pa.getFloat( "CROP_CZMIN" );
                        if ( pa.getState( "CROP_CXMAX" ) ) crop_tmp[3] = pa.getFloat( "CROP_CXMAX" );
                        if ( pa.getState( "CROP_CYMAX" ) ) crop_tmp[4] = pa.getFloat( "CROP_CYMAX" );
                        if ( pa.getState( "CROP_CZMAX" ) ) crop_tmp[5] = pa.getFloat( "CROP_CZMAX" );
                        param.crop.set_enable( pa.getInt( "CROP_TYPE" ) );
                        param.crop.set( crop_tmp );
                        // printf("min(%f,%f,%f)\n", crop_tmp[0], crop_tmp[1], crop_tmp[2]);
                        // printf("max(%f,%f,%f)\n", crop_tmp[3], crop_tmp[4], crop_tmp[5]);
                    }
                    break;

                    case 2:
                    {
                        float crop_tmp[6];
                        crop_tmp[0] = ( fil.total_minObjectCoord.x() + fil.total_maxObjectCoord.x() ) / 2.0;
                        crop_tmp[1] = ( fil.total_minObjectCoord.y() + fil.total_maxObjectCoord.y() ) / 2.0;
                        crop_tmp[2] = ( fil.total_minObjectCoord.z() + fil.total_maxObjectCoord.z() ) / 2.0;
                        float tx = fil.total_maxObjectCoord.x() - crop_tmp[0];
                        float ty = fil.total_maxObjectCoord.y() - crop_tmp[1];
                        float tz = fil.total_maxObjectCoord.z() - crop_tmp[2];
                        crop_tmp[3] = sqrt( tx * tx + ty * ty + tz * tz );
                        if ( pa.getState( "CROP_SCENTERX" ) ) crop_tmp[0] = pa.getFloat( "CROP_SCENTERX" );
                        if ( pa.getState( "CROP_SCENTERY" ) ) crop_tmp[1] = pa.getFloat( "CROP_SCENTERY" );
                        if ( pa.getState( "CROP_SCENTERZ" ) ) crop_tmp[2] = pa.getFloat( "CROP_SCENTERZ" );
                        if ( pa.getState( "CROP_SRADIUS" ) ) crop_tmp[3] = pa.getFloat( "CROP_SRADIUS" );
                        crop_tmp[4] = 0.0;
                        crop_tmp[5] = 0.0;
                        param.crop.set_enable( pa.getInt( "CROP_TYPE" ) );;
                        param.crop.set( crop_tmp );
                        // printf("(%f,%f,%f) %f\n", crop_tmp[0], crop_tmp[1], crop_tmp[2], crop_tmp[3]);
                    }
                    break;

                    case 3:         // XY
                    {
                        float crop_tmp[6];
                        crop_tmp[0] = ( fil.total_minObjectCoord.x() + fil.total_maxObjectCoord.x() ) / 2.0;
                        crop_tmp[1] = ( fil.total_minObjectCoord.y() + fil.total_maxObjectCoord.y() ) / 2.0;
                        crop_tmp[2] = fil.total_minObjectCoord.z();
                        float tx = fil.total_maxObjectCoord.x() - crop_tmp[0];
                        float ty = fil.total_maxObjectCoord.y() - crop_tmp[1];
                        crop_tmp[3] = sqrt( tx * tx + ty * ty );
                        crop_tmp[4] = fil.total_maxObjectCoord.z() - fil.total_minObjectCoord.z();
                        if ( pa.getState( "CROP_PCENTERX" ) ) crop_tmp[0] = pa.getFloat( "CROP_PCENTERX" );
                        if ( pa.getState( "CROP_PCENTERY" ) ) crop_tmp[1] = pa.getFloat( "CROP_PCENTERY" );
                        if ( pa.getState( "CROP_PCENTERZ" ) ) crop_tmp[2] = pa.getFloat( "CROP_PCENTERZ" );
                        if ( pa.getState( "CROP_PRADIUS" ) ) crop_tmp[3] = pa.getFloat( "CROP_PRADIUS" );
                        if ( pa.getState( "CROP_PHEIGHT" ) ) crop_tmp[4] = pa.getFloat( "CROP_PHEIGHT" );
                        crop_tmp[5] = 0.0;
                        param.crop.set_enable( pa.getInt( "CROP_TYPE" ) );;
                        param.crop.set( crop_tmp );
                        // printf("xy (%f,%f,%f) %f, %f\n", crop_tmp[0], crop_tmp[1], crop_tmp[2],
                        //        crop_tmp[3], crop_tmp[4]);
                    }
                    break;

                    case 4:         // YZ
                    {
                        float crop_tmp[6];
                        crop_tmp[0] = fil.total_minObjectCoord.x();
                        crop_tmp[1] = ( fil.total_minObjectCoord.y() + fil.total_maxObjectCoord.y() ) / 2.0;
                        crop_tmp[2] = ( fil.total_minObjectCoord.z() + fil.total_maxObjectCoord.z() ) / 2.0;
                        float ty = fil.total_maxObjectCoord.y() - crop_tmp[1];
                        float tz = fil.total_maxObjectCoord.z() - crop_tmp[2];
                        crop_tmp[3] = sqrt( ty * ty + tz * tz );
                        crop_tmp[4] = fil.total_maxObjectCoord.x() - fil.total_minObjectCoord.x();
                        if ( pa.getState( "CROP_PCENTERX" ) ) crop_tmp[0] = pa.getFloat( "CROP_PCENTERX" );
                        if ( pa.getState( "CROP_PCENTERY" ) ) crop_tmp[1] = pa.getFloat( "CROP_PCENTERY" );
                        if ( pa.getState( "CROP_PCENTERZ" ) ) crop_tmp[2] = pa.getFloat( "CROP_PCENTERZ" );
                        if ( pa.getState( "CROP_PRADIUS" ) ) crop_tmp[3] = pa.getFloat( "CROP_PRADIUS" );
                        if ( pa.getState( "CROP_PHEIGHT" ) ) crop_tmp[4] = pa.getFloat( "CROP_PHEIGHT" );
                        crop_tmp[5] = 0.0;
                        param.crop.set_enable( pa.getInt( "CROP_TYPE" ) );;
                        param.crop.set( crop_tmp );
                        // printf("yz (%f,%f,%f) %f, %f\n", crop_tmp[0], crop_tmp[1], crop_tmp[2],
                        //        crop_tmp[3], crop_tmp[4]);
                    }
                    break;

                    case 5:         // XZ
                    {
                        float crop_tmp[6];
                        crop_tmp[0] = ( fil.total_minObjectCoord.x() + fil.total_maxObjectCoord.x() ) / 2.0;
                        crop_tmp[1] = fil.total_minObjectCoord.y();
                        crop_tmp[2] = ( fil.total_minObjectCoord.z() + fil.total_maxObjectCoord.z() ) / 2.0;
                        float tx = fil.total_maxObjectCoord.x() - crop_tmp[0];
                        float tz = fil.total_maxObjectCoord.z() - crop_tmp[2];
                        crop_tmp[3] = sqrt( tx * tx + tz * tz );
                        crop_tmp[4] = fil.total_maxObjectCoord.y() - fil.total_minObjectCoord.y();
                        if ( pa.getState( "CROP_PCENTERX" ) ) crop_tmp[0] = pa.getFloat( "CROP_PCENTERX" );
                        if ( pa.getState( "CROP_PCENTERY" ) ) crop_tmp[1] = pa.getFloat( "CROP_PCENTERY" );
                        if ( pa.getState( "CROP_PCENTERZ" ) ) crop_tmp[2] = pa.getFloat( "CROP_PCENTERZ" );
                        if ( pa.getState( "CROP_PRADIUS" ) ) crop_tmp[3] = pa.getFloat( "CROP_PRADIUS" );
                        if ( pa.getState( "CROP_PHEIGHT" ) ) crop_tmp[4] = pa.getFloat( "CROP_PHEIGHT" );
                        crop_tmp[5] = 0.0;
                        param.crop.set_enable( pa.getInt( "CROP_TYPE" ) );;
                        param.crop.set( crop_tmp );
                        // printf("xz (%f,%f,%f) %f, %f\n", crop_tmp[0], crop_tmp[1], crop_tmp[2],
                        //        crop_tmp[3], crop_tmp[4]);
                    }
                    break;

                    default:
                        break;
                    }
                }
                if ( pa.getState( "SERVER_LATENCY" ) ) param.latency_threshold = ( double )pa.getFloat( "SERVER_LATENCY" );
                if ( pa.getState( "SERVER_JOB_NUMBER" ) ) param.jid_pack_size = pa.getInt( "SERVER_JOB_NUMBER" );

                transfunc_creator.setParamInfo( pa );
            }
        } // end of if( param.param_file != "" )
        TIMER_END( 13 );

        transfunc_creator.setAsisTransferFunction ( param.transfer_function );
        if ( param.hasOption( "t" ) )
            transfunc_creator.setTransferFunction ( "t1", param.transfer_function );
        param.transfunc_synthesizer = transfunc_creator.create();
        param.transfunc_synthesizer->optimize2();

        camera.setWindowSize( param.windowWidth, param.windowHeight );
        TIMER_STA( 15 );
        //kawamura2 : It calculates subpixel level from particle limit.
        param.sampling_step = CalculateSamplingStep( fil );
        param.subpixel_level = CalculateSubpixelLevel( param, fil, camera );
        TIMER_END( 15 );

        TIMER_END( 11 );

        std::cout << "subpixel level = " << param.subpixel_level << std::endl;
        std::cout << "sampling_step = " << param.sampling_step << std::endl;

        if ( param.batch_fjoin )
        {
            //
            // batch fjoin mode: integrate into a single file for each time step
            //
            pbvr::PointObject joined_obj;
            int32_t stp;
#if defined(CPU_VER)
            for ( stp = fil.total_staSteps; stp <= fil.total_endSteps; stp++ )
            {
                joined_obj.clear();
                TIMER_STA( 14 );
                if ( param.crop.isenabled() )
                {
                    jd.Initialize( stp, stp, fil.total_numSubVolumes,
                                   fil.total_minSubVolumeCoord,
                                   fil.total_maxSubVolumeCoord,
                                   param.latency_threshold, param.jid_pack_size,
                                   param.crop.get_min_coord(),
                                   param.crop.get_max_coord() );
                }
                else
                {
                    jd.Initialize( stp, stp, fil.total_numSubVolumes,
                                   fil.total_minSubVolumeCoord,
                                   fil.total_maxSubVolumeCoord,
                                   param.latency_threshold, param.jid_pack_size );
                }
                TIMER_END( 14 );

                TIMER_STA( 20 );
                while ( jd.DispatchNext( wid, &st, &vl ) )
                {
                    TIMER_STA( 90 );

                    int xvl, fidx;
                    fidx = fil.GetFileIdx( vl, xvl );
                    FilterInfo& fi = fil.m_list[fidx];

                    std::stringstream suffix, suffix2;
                    suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << st
                           << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( xvl + 1 )
                           << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;
                    //param.input_data = param.input_data_base + suffix.str() + ".kvsml";
                    kvs::File ifpx( fi.filePath );
                    param.input_data = ifpx.pathName() + ifpx.Separator()
                                       + ifpx.baseName() + suffix.str() + ".kvsml";

                    int timeStep = 1;
                    if ( fi.fileType == 1 || fi.fileType == 2 ) // filetype: gathered subvolume or gathered timestep
                    {
                        TIMER_STA( 99 );
                        object = point_creator_lst[fidx].run( param, camera, timeStep, st, xvl );
                        TIMER_END( 99 );
                    }
                    else            // filetype: kvsml
                    {
                        TIMER_STA( 99 );
                        object = point_creator_lst[fidx].run( param, camera, timeStep );
                        TIMER_END( 99 );
                    }

                    if ( object )
                    {
                        int nvertices = object->coords().size() / 3;
                        if ( nvertices > 0 )
                        {
                            joined_obj.add( *object );
                        }
                    }

                    TIMER_END( 90 );
                } // END OF WHILE(DispatchNext)Job
                TIMER_END( 20 );

                // WRITE joined_obj
                std::stringstream suffix2;
                suffix2 << '_' << std::setw( 5 ) << std::setfill( '0' ) << stp
                        << '_' << std::setw( 7 ) << std::setfill( '0' ) << 1
                        << '_' << std::setw( 7 ) << std::setfill( '0' ) << 1;
                std::string basename = output + suffix2.str();
                joined_obj.setSize( static_cast<kvs::Real32>( param.subpixel_level ) );
                KVSMLObjectPointWriter( &joined_obj, basename );

            } // end of for(stp)

#else // ! CPU_VER

            for ( stp = fil.total_staSteps; stp <= fil.total_endSteps; stp++ )
            {
                joined_obj.clear();
                TIMER_STA( 14 );
                if ( param.crop.isenabled() )
                {
                    jd.Initialize( stp, stp, fil.total_numSubVolumes,
                                   fil.total_minSubVolumeCoord,
                                   fil.total_maxSubVolumeCoord,
                                   param.latency_threshold, param.jid_pack_size,
                                   param.crop.get_min_coord(),
                                   param.crop.get_max_coord() );
                }
                else
                {
                    jd.Initialize( stp, stp, fil.total_numSubVolumes,
                                   fil.total_minSubVolumeCoord,
                                   fil.total_maxSubVolumeCoord,
                                   param.latency_threshold, param.jid_pack_size );
                }
                TIMER_END( 14 );

                TIMER_STA( 20 );
                while ( jd.DispatchNext( wid, &st, &vl ) )
                {

                    if ( rank > 0 )
                    {
                        //--------------------- WORKER --------------------
                        TIMER_STA( 90 );

                        int xvl, fidx;
                        fidx = fil.GetFileIdx( vl, xvl );
                        FilterInfo& fi = fil.m_list[fidx];

                        std::stringstream suffix, suffix2;
                        suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << st
                               << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( xvl + 1 )
                               << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;
                        //param.input_data = param.input_data_base + suffix.str() + ".kvsml";
                        kvs::File ifpx( fi.filePath );
                        param.input_data = ifpx.pathName() + ifpx.Separator()
                                           + ifpx.baseName() + suffix.str() + ".kvsml";

                        int timeStep = 1;
                        if ( fi.fileType == 1 || fi.fileType == 2 ) // filetype: gathered subvolume or gathered timestep
                        {
                            TIMER_STA( 99 );
                            object = point_creator_lst[fidx].run( param, camera, timeStep, st, xvl );
                            TIMER_END( 99 );
                        }
                        else            // filetype: kvsml
                        {
                            TIMER_STA( 99 );
                            object = point_creator_lst[fidx].run( param, camera, timeStep );
                            TIMER_END( 99 );
                        }
                        TIMER_STA( 400 );

                        VariableRange* p_vr = &param.transfunc_synthesizer->variableRange();
                        jc.JobCollect( object, p_vr, &wid );

                        TIMER_END( 400 );
                        TIMER_END( 90 );
                    }
                    else
                    {
                        // ---------------------- MASTER ---------------------
                        TIMER_STA( 50 );
                        pbvr::PointObject obj;
                        TIMER_STA( 55 );

                        VariableRange* p_vr = &param.transfunc_synthesizer->variableRange();
                        jc.JobCollect( &obj, p_vr, &wid );

                        int nvertices = obj.coords().size() / 3;
                        if ( nvertices > 0 )
                        {
                            joined_obj.add( obj );
                        }

                        jd.setWorkerStatus( wid, 1 );

                        TIMER_END( 55 );
                        TIMER_END( 50 );
                    }

                } // END OF WHILE(DispatchNext)Job
                TIMER_END( 20 );

                if ( rank == 0 )
                {
                    // WRITE joined_obj
                    std::stringstream suffix2;
                    suffix2 << '_' << std::setw( 5 ) << std::setfill( '0' ) << stp
                            << '_' << std::setw( 7 ) << std::setfill( '0' ) << 1
                            << '_' << std::setw( 7 ) << std::setfill( '0' ) << 1;
                    std::string basename = output + suffix2.str();
                    joined_obj.setSize( static_cast<kvs::Real32>( param.subpixel_level ) );
                    KVSMLObjectPointWriter( &joined_obj, basename );
                }
            } // end of for(stp)

#endif // ! CPU_VER
        }
        else
        {
            //
            // NOT fjoin mode: file will be created apart in each subvolume
            //
            TIMER_STA( 14 );
            if ( param.crop.isenabled() )
            {
                jd.Initialize( fil.total_staSteps, fil.total_endSteps, fil.total_numSubVolumes,
                               fil.total_minSubVolumeCoord,
                               fil.total_maxSubVolumeCoord,
                               param.latency_threshold, param.jid_pack_size,
                               param.crop.get_min_coord(),
                               param.crop.get_max_coord() );
            }
            else
            {
                jd.Initialize( fil.total_staSteps, fil.total_endSteps, fil.total_numSubVolumes,
                               fil.total_minSubVolumeCoord,
                               fil.total_maxSubVolumeCoord,
                               param.latency_threshold, param.jid_pack_size );
            }
            TIMER_END( 14 );

            TIMER_STA( 20 );
            while ( jd.DispatchNext( wid, &st, &vl ) )
            {

#ifdef CPU_VER
                if ( rank == 0 )
#else
                if ( rank > 0 )
#endif
                {
                    //--------------------- WORKER --------------------
                    TIMER_STA( 90 );

                    int xvl, fidx;
                    fidx = fil.GetFileIdx( vl, xvl );
                    FilterInfo& fi = fil.m_list[fidx];

                    std::stringstream suffix, suffix2;
                    suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << st
                           << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( xvl + 1 )
                           << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;
                    suffix2 << '_' << std::setw( 5 ) << std::setfill( '0' ) << st
                            << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( vl + 1 )
                            << '_' << std::setw( 7 ) << std::setfill( '0' ) << fil.total_numSubVolumes;
                    //param.input_data = param.input_data_base + suffix.str() + ".kvsml";
                    kvs::File ifpx( fi.filePath );
                    param.input_data = ifpx.pathName() + ifpx.Separator()
                                       + ifpx.baseName() + suffix.str() + ".kvsml";
                    std::string basename = output + suffix2.str();

                    int timeStep = 1;
                    if ( fi.fileType == 1 || fi.fileType == 2 ) // filetype: gathered subvolume or gathered timestep
                    {
                        TIMER_STA( 99 );
                        object = point_creator_lst[fidx].run( param, camera, timeStep, st, xvl );
                        TIMER_END( 99 );
                    }
                    else            // filetype: kvsml
                    {
                        TIMER_STA( 99 );
                        object = point_creator_lst[fidx].run( param, camera, timeStep );
                        TIMER_END( 99 );
                    }
                    TIMER_STA( 80 );
                    //サイズの代わりにサブピクセルレベルを代入.
                    object->setSize( static_cast<kvs::Real32>( param.subpixel_level ) );
                    KVSMLObjectPointWriter( object, basename );
                    TIMER_END( 80 );
                    TIMER_STA( 400 );

#ifndef CPU_VER
                    VariableRange* p_vr = &param.transfunc_synthesizer->variableRange();
                    jc.JobCollect( object, p_vr, &wid );
#endif

                    TIMER_END( 400 );
                    TIMER_END( 90 );
                }
                else
                {
                    // ---------------------- MASTER ---------------------
                    TIMER_STA( 50 );
                    pbvr::PointObject obj;
                    TIMER_STA( 55 );
#ifndef CPU_VER
                    VariableRange* p_vr = &param.transfunc_synthesizer->variableRange();
                    jc.JobCollect( &obj, p_vr, &wid );
                    jd.setWorkerStatus( wid, 1 );
#endif
                    TIMER_END( 55 );
                    TIMER_END( 50 );
                }

            } // END OF WHILE(DispatchNext)Job
            TIMER_END( 20 );
        } // end of ! batch_fjoin

        delete param.transfunc_synthesizer;
        delete object;

        TIMER_END( 2 );
    } // end of batch mode

    else  //=================== client-server mode ===================
    {

        char* buf;
        int bsz = 0;
        JobDispatcher jd;
#ifndef CPU_VER
        JobCollector  jc( &jd );
#endif
        int st, vl, wid = 0;

        int c_bins_size = 0;
        int o_bins_size = 0;
        kvs::UInt64* tmp_c_bins;
        kvs::UInt64* tmp_o_bins;

        if ( rank > 0 )
        {
            //--------------------- WORKER --------------------
            jpv::ParticleTransferClientMessage clntMes;
            clntMes.camera = new kvs::Camera();

            bool loop = true;

            while ( loop )
            {
                static int timer_count = 0;

                // recv cltMes from process 0 >>
#ifndef CPU_VER
                MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD );
#endif
                if ( bsz < 0 )
                {
                    loop = false;
                    break; // terminate server
                }
                buf = new char[bsz];
#ifndef CPU_VER
                MPI_Bcast( buf, bsz, MPI_BYTE, 0, MPI_COMM_WORLD );
#endif
                clntMes.unpack( buf );
                delete[] buf;
                std::cout << "Rank " << rank << ": Recv Client Message" << std::endl;
                // recv cltMes from process 0 <<

                if ( clntMes.timeParam == 0 )
                {
                }
                else if ( clntMes.timeParam == 1 )
                {
                }
                else
                {
                    timer_count++;
//                  param.transfer_function = pbvr::TransferFunction(); // *( clntMes.transfunc );
                    param.sampling_method = clntMes.samplingMethod;
                    param.component_Id = clntMes.renderingId;
                    param.crop.set_enable( clntMes.enable_crop_region );
                    param.crop.set( clntMes.crop_region );
                    param.input_data_base = clntMes.inputDir;
                    param.particle_limit = clntMes.particle_limit;
                    param.particle_density = clntMes.particle_density;

                    std::string pfifile, pflfile;
                    if ( param.input_data_base.substr( param.input_data_base.size() - 3 ) == "pfl" )
                    {
                        pflfile = param.input_data_base;
                        kvs::File pfl( pflfile );
                        if ( pfl.isExisted() )
                        {
                            fil.LoadPFL( pflfile );
                        }
                    }
                    else
                    {
                        pfifile = param.input_data_base + ".pfi";
                        kvs::File pfi( pfifile );
                        pflfile = param.input_data_base + ".pfl";
                        kvs::File pfl( pflfile );
                        if ( pfl.isExisted() )
                        {
                            fil.LoadPFL( pflfile );
                        }
                        else if ( pfi.isExisted() )
                        {
                            fil.LoadPFL( pfifile );
                        }
                    }

                    //point_creator.setFilterInfo( fi );
                    point_creator_lst.clear();
                    for ( int idx = 0; idx < fil.m_list.size(); idx++ )
                    {
                        PointObjectCreator point_creator;
                        if ( param.GT5D == true ) point_creator.setGT5D();
                        point_creator.setFilterInfo( fil.m_list[idx] );
                        point_creator_lst.push_back( point_creator );
                    }

                    transfunc_creator.setFilterInfo( fil.m_list[0] );
                    transfunc_creator.setProtocol( clntMes );
                    transfunc_creator.setAsisTransferFunction( param.transfer_function );
                    param.transfunc_synthesizer = transfunc_creator.create();
                    param.transfunc_synthesizer->optimize2();

                    if ( !param.hasOption( "L" ) ) param.latency_threshold = -1.0;
                    if ( param.crop.isenabled() )
                    {
                        jd.Initialize( clntMes.step, clntMes.step, fil.total_numSubVolumes,
                                       fil.total_minSubVolumeCoord,
                                       fil.total_maxSubVolumeCoord,
                                       param.latency_threshold, param.jid_pack_size,
                                       param.crop.get_min_coord(),
                                       param.crop.get_max_coord() );
                    }
                    else
                    {
                        jd.Initialize( clntMes.step, clntMes.step, fil.total_numSubVolumes,
                                       fil.total_minSubVolumeCoord,
                                       fil.total_maxSubVolumeCoord,
                                       param.latency_threshold, param.jid_pack_size );
                    }

                    param.sampling_step = CalculateSamplingStep( fil );
                    param.subpixel_level = CalculateSubpixelLevel( param, fil, camera );
                    param.particle_limit_pre = param.particle_limit;

                    int tf_count = TF_COUNT;
                    int c_nbins = DEFAULT_NBINS;
                    int o_nbins = DEFAULT_NBINS;

                    c_bins_size = 0;
                    o_bins_size = 0;

                    for ( int tf = 0; tf < tf_count; tf++ )
                    {
                        c_bins_size += c_nbins;
                        o_bins_size += o_nbins;
                    }

                    tmp_c_bins = new kvs::UInt64[c_bins_size];
                    tmp_o_bins = new kvs::UInt64[o_bins_size];

                    for ( int tf = 0; tf < c_bins_size; tf++ )
                    {
                        tmp_c_bins[tf] = 0;
                    }

                    for ( int tf = 0; tf < o_bins_size; tf++ )
                    {
                        tmp_o_bins[tf] = 0;
                    }

                    while ( jd.DispatchNext( wid, &st, &vl ) )
                    {
                        int xvl, fidx;
                        fidx = fil.GetFileIdx( vl, xvl );
                        FilterInfo& fi = fil.m_list[fidx];

                        std::stringstream suffix;
                        suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << ( st )
                               << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( xvl + 1 )
                               << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;
                        //param.input_data = param.input_data_base + suffix.str() + ".kvsml";
                        kvs::File ifpx( fi.filePath );
                        param.input_data = ifpx.pathName() + ifpx.Separator()
                                           + ifpx.baseName() + suffix.str() + ".kvsml";
                        int timeStep = 1;
                        if ( fi.fileType == 1 || fi.fileType == 2 ) // filetype: gathered subvolume or gathered timestep
                        {
                            object = point_creator_lst[fidx].run( param, *clntMes.camera, timeStep, st, xvl );

                        }
                        else     // filetype: kvsml
                        {
                            object = point_creator_lst[fidx].run( param, *clntMes.camera, timeStep );
                        }
#ifndef CPU_VER
                        VariableRange* p_vr = &param.transfunc_synthesizer->variableRange();
                        jc.JobCollect( object, p_vr, &wid );
#endif

                        int c_count = 0;
                        int o_count = 0;

                        for ( int tf = 0; tf < TF_COUNT; tf++ )
                        {
                            c_nbins = object->getColorHistogram()[ tf ].nbins();
                            for ( int res = 0; res < c_nbins; res++ )
                            {
                                tmp_c_bins[c_count] += static_cast<kvs::UInt64>( object->getColorHistogram()[ tf ][res] );
                                c_count++;
                            }
                        }

                        for ( int tf = 0; tf < TF_COUNT; tf++ )
                        {
                            o_nbins = object->getOpacityHistogram()[ tf ].nbins();
                            for ( int res = 0; res < o_nbins; res++ )
                            {
                                tmp_o_bins[o_count] += static_cast<kvs::UInt64>( object->getOpacityHistogram()[ tf ][res] );
                                o_count++;
                            }
                        }

                    } // end of while(DispatchNext)
#ifndef CPU_VER

                    MPI_Allreduce( MPI_IN_PLACE, tmp_c_bins, c_bins_size, MPI_UNSIGNED_LONG, MPI_SUM , MPI_COMM_WORLD );
                    MPI_Allreduce( MPI_IN_PLACE, tmp_o_bins, o_bins_size, MPI_UNSIGNED_LONG, MPI_SUM , MPI_COMM_WORLD );
                    delete[] tmp_c_bins;
                    delete[] tmp_o_bins;
#endif
                    if ( timer_count == TIMER_COUNT_NUM )
                    {
                        TIMER_END( 1 );
                        TIMER_FIN();
                    }
                    delete param.transfunc_synthesizer;
                }
            } // end of while(loop)
            delete clntMes.camera;

            delete object;

        }
        else                    // rank == 0
        {
            //--------------------- MASTER --------------------
            int ptss;

            assert( jpv::ParticleTransferUtils::isLittleEndian() );
            timer.start();
            TIMER_STA( 10 );
            //XX//if( param.GT5D == true ) point_creator.setGT5D();
            TIMER_END( 10 );
            timer.stop();
            std::cout << "first reading time[ms]:" << timer.msec() << std::endl;

            jpv::ParticleTransferServer pts;
            ptss = pts.initServer( param.port );
            if ( ptss == -1 )
            {
                bsz = -1;
#ifndef CPU_VER
                MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD ); // termination message
#endif
            }
            else
            {
                pts.acceptServer();
                jpv::ParticleTransferServerMessage servMes;
                jpv::ParticleTransferClientMessage clntMes;

                pts.recvMessage( clntMes );
                param.input_data_base = clntMes.inputDir;

                std::string pflfile, pfifile;
                if ( param.input_data_base.substr( param.input_data_base.size() - 3 ) == "pfl" )
                {
                    pflfile = param.input_data_base;
                    kvs::File pfl( pflfile );
                    if ( pfl.isExisted() )
                    {
                        fil.LoadPFL( pflfile );
                    }
                }
                else
                {
                    pfifile = param.input_data_base + ".pfi";
                    kvs::File pfi( pfifile );
                    pflfile = param.input_data_base + ".pfl";
                    kvs::File pfl( pflfile );
                    if ( pfl.isExisted() )
                    {
                        fil.LoadPFL( pflfile );
                    }
                    else if ( pfi.isExisted() )
                    {
                        fil.LoadPFL( pfifile );
                    }
                }

                if ( fil.m_list.size() > 0 )
                {
                    point_creator_lst.clear();
                    for ( int idx = 0; idx < fil.m_list.size(); idx++ )
                    {
                        PointObjectCreator point_creator;
                        if ( param.GT5D == true ) point_creator.setGT5D();
                        point_creator.setFilterInfo( fil.m_list[idx] );
                        point_creator_lst.push_back( point_creator );
                    }

                    transfunc_creator.setFilterInfo( fil.m_list[0] );

                    std::cout << " start step = "         << fil.total_staSteps
                              << " end step = "           << fil.total_endSteps
                              << " time step = "          << fil.total_numSteps
                              << " subvolume division = " << fil.total_numSubVolumes
                              << std::endl;

                }
                else
                {
                    std::cerr << "Error: pfifile (" << pfifile << ") nor pflfile ("
                              << pflfile << ") doesn't exist" << std::endl;
                    bsz = -1;
#ifndef CPU_VER
                    MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD ); // termination message
#endif
#ifdef KMATH
                    km_random.finalize();
#endif
#ifndef CPU_VER
                    MPI::Finalize();
#endif
                    return 0;
                }

                transfunc_creator.setProtocol( clntMes );
                TransferFunctionSynthesizer* tfs = transfunc_creator.create();
                VariableRange range = RangeEstimater::Estimation( 0, fil.m_list[0], *tfs );
                delete tfs;

                strncpy( servMes.header, "JPTP /1.0 000 OK\r\n", 18 );
                servMes.numParticle = 0;
                servMes.tf_count = 0;
                servMes.numVolDiv = fil.total_numSubVolumes;
                servMes.timeStep = fil.total_staSteps;
                servMes.staStep = fil.total_staSteps;
                servMes.endStep = fil.total_endSteps;
                servMes.numStep = fil.total_numSteps;
                servMes.minObjectCoord[0] = fil.total_minObjectCoord[0];
                servMes.minObjectCoord[1] = fil.total_minObjectCoord[1];
                servMes.minObjectCoord[2] = fil.total_minObjectCoord[2];
                servMes.maxObjectCoord[0] = fil.total_maxObjectCoord[0];
                servMes.maxObjectCoord[1] = fil.total_maxObjectCoord[1];
                servMes.maxObjectCoord[2] = fil.total_maxObjectCoord[2];
                servMes.minValue = fil.total_minValue;
                servMes.maxValue = fil.total_maxValue;
                servMes.numNodes = fil.total_numNodes;
                servMes.numElements = fil.total_numElements;
                servMes.elemType = fil.m_list[0].elemType;
                servMes.fileType = fil.m_list[0].fileType;
                servMes.numIngredients = fil.m_list[0].numIngredients;
                servMes.varRange = range;
                servMes.flag_send_bins = 0;

                servMes.messageSize = servMes.byteSize();
                pts.sendMessage( servMes );
            }

            jpv::ParticleTransferServerMessage servMes;
            jpv::ParticleTransferClientMessage clntMes;
            clntMes.camera = new kvs::Camera();

            while ( ( ptss != -1 ) && ( pts.good() ) )
            {
                static int timer_count = 0;

                ptss = pts.recvMessage( clntMes );
                if ( ptss == -1 ) break;
                /* 140319 for client stop by Ctrl+c */
                signal( SIGABRT, sig_handler );
                signal( SIGTERM, sig_handler );
                signal( SIGINT, sig_handler ); /* SIGINT is invalid here, because mpiexec uses it. */
//              signal( SIGSEGV, sig_handler );
                if ( clntMes.step > fil.total_endSteps )
                {
                    clntMes.step = fil.total_endSteps;
                }
                else if ( clntMes.step < fil.total_staSteps )
                {
                    clntMes.step = fil.total_staSteps;
                }

                if ( SigServer )
                {
                    clntMes.initParam = -2;
                    std::cout << "*** SigServer" << clntMes.initParam << std::endl;
                }
                else
                {
                    /* 140319 for client stop by Ctrl+c */
                    if ( clntMes.initParam != -3 )
                    {
                        clntMes.inputDir = param.input_data_base;
                    }
                }

                std::cout << "Recieve message initParam = " << clntMes.initParam << std::endl;
                if ( clntMes.initParam == -1 )
                {

                    strncpy( servMes.header, "JPTP /1.0 899 OK\r\n", 18 );
                    servMes.numParticle = 0;
                    servMes.flag_send_bins = 0;

                    servMes.messageSize = servMes.byteSize();
                    pts.sendMessage( servMes );
                    pts.disconnect();

                    pts.acceptServer();
                }
                else if ( clntMes.initParam == -2 )
                {

                    strncpy( servMes.header, "JPTP /1.0 999 OK\r\n", 18 );
                    servMes.numParticle = 0;
                    servMes.flag_send_bins = 0;

                    servMes.messageSize = servMes.byteSize();
                    pts.sendMessage( servMes );
                    break;
                }
                else if ( clntMes.initParam == -3 ) // change PFI file.
                {
                    param.input_data_base = clntMes.inputDir;

                    std::string pfifile = param.input_data_base + ".pfi";
                    kvs::File pfi( pfifile );
                    std::string pflfile = param.input_data_base + ".pfl";
                    kvs::File pfl( pflfile );
                    if ( pfl.isExisted() )
                    {
                        fil.LoadPFL( pflfile );
                    }
                    else if ( pfi.isExisted() )
                    {
                        fil.LoadPFL( pfifile );
                    }

                    if ( fil.m_list.size() > 0 )
                    {
                        point_creator_lst.clear();
                        for ( int idx = 0; idx < fil.m_list.size(); idx++ )
                        {
                            PointObjectCreator point_creator;
                            if ( param.GT5D == true ) point_creator.setGT5D();
                            point_creator.setFilterInfo( fil.m_list[idx] );
                            point_creator_lst.push_back( point_creator );
                        }

                        transfunc_creator.setFilterInfo( fil.m_list[0] );

                        std::cout << " time step = "          << fil.total_numSteps
                                  << " subvolume division = " << fil.total_numSubVolumes
                                  << std::endl;

                    }
                    else
                    {
                        if ( rank == 0 ) std::cerr << "Error: pfifile doesn't exist" << std::endl;
                        bsz = -1;
#ifndef CPU_VER
                        MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD ); // termination message
#endif
#ifdef KMATH
                        km_random.finalize();
#endif
#ifndef CPU_VER
                        MPI::Finalize();
#endif
                        return 0;
                    }

                    transfunc_creator.setProtocol( clntMes );
                    TransferFunctionSynthesizer* tfs = transfunc_creator.create();
                    VariableRange range = RangeEstimater::Estimation( 0, fil.m_list[0], *tfs );
                    delete tfs;

                    strncpy( servMes.header, "JPTP /1.0 000 OK\r\n", 18 );
                    servMes.numParticle = 0;
                    servMes.numVolDiv = fil.total_numSubVolumes;
                    servMes.timeStep = fil.total_staSteps;
                    servMes.staStep = fil.total_staSteps;
                    servMes.endStep = fil.total_endSteps;
                    servMes.numStep = fil.total_endSteps;
                    servMes.minObjectCoord[0] = fil.total_minObjectCoord[0];
                    servMes.minObjectCoord[1] = fil.total_minObjectCoord[1];
                    servMes.minObjectCoord[2] = fil.total_minObjectCoord[2];
                    servMes.maxObjectCoord[0] = fil.total_maxObjectCoord[0];
                    servMes.maxObjectCoord[1] = fil.total_maxObjectCoord[1];
                    servMes.maxObjectCoord[2] = fil.total_maxObjectCoord[2];
                    servMes.minValue = fil.total_minValue;
                    servMes.maxValue = fil.total_maxValue;
                    servMes.numNodes = fil.total_numNodes;
                    servMes.numElements = fil.total_numElements;
                    servMes.elemType = fil.m_list[0].elemType;
                    servMes.fileType = fil.m_list[0].fileType;
                    servMes.numIngredients = fil.m_list[0].numIngredients;
                    servMes.varRange = range;
                    servMes.flag_send_bins = 0;

                    servMes.messageSize = servMes.byteSize();
                    pts.sendMessage( servMes );
                } // end of change PFI
                else
                {
                    timer_count++;
                    if ( timer_count <= TIMER_COUNT_NUM )
                    {
                        TIMER_STA( 461 );
                    }

                    // send cltMes to all worker process >>
                    bsz = clntMes.byteSize();
#ifndef CPU_VER
                    MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD );
#endif
                    buf = new char[bsz];
                    clntMes.pack( buf );
#ifndef CPU_VER
                    MPI_Bcast( buf, bsz, MPI_BYTE, 0, MPI_COMM_WORLD );
#endif
                    delete[] buf;
                    // send cltMes to all worker process <<

                    std::cout << "initParam = " << clntMes.initParam << std::endl;
                    if ( clntMes.initParam == 1 )
                    {

                        std::cout << "sampling method = " << clntMes.samplingMethod << std::endl;
                        std::cout << "subpixel level = " << clntMes.subPixelLevel << std::endl;
                        std::cout << "repeat level = " << clntMes.repeatLevel << std::endl;
                    }
                    std::cout << "timeParam = " << clntMes.timeParam << std::endl;

                    if ( clntMes.timeParam == 0 )
                    {
                        std::cout << "memorySize = " << clntMes.memorySize << std::endl;
                    }
                    else if ( clntMes.timeParam == 1 )
                    {
                        std::cout << "beginTime = " << clntMes.beginTime << std::endl;
                        std::cout << "endTime = " << clntMes.endTime << std::endl;
                        std::cout << "memorySize = " << clntMes.memorySize << std::endl;
                    }
                    else if ( clntMes.timeParam == 2 )
                    {
                        std::cout << "step = " << clntMes.step << std::endl;
                    }
                    std::cout << "transParam = " << clntMes.transParam << std::endl;
                    if ( clntMes.transParam == 1 )
                    {
                        std::cout << "levelIndex = " << clntMes.levelIndex << std::endl;
                    }
                    if ( clntMes.timeParam == 0 )
                    {
                        strncpy( servMes.header, "JPTP /1.0 130 OK\r\n", 18 );
                        servMes.timeStep = clntMes.step;
                        servMes.repeatLevel = clntMes.repeatLevel;
                        servMes.levelIndex = clntMes.levelIndex;
                        servMes.numParticle = 0;
                        servMes.flag_send_bins = 0;

                        servMes.messageSize = servMes.byteSize();
                        pts.sendMessage( servMes );
                    }
                    else if ( clntMes.timeParam == 1 )
                    {

                        strncpy( servMes.header, "JPTP /1.0 130 OK\r\n", 18 );
                        servMes.timeStep = clntMes.step;
                        servMes.repeatLevel = clntMes.repeatLevel;
                        servMes.levelIndex = clntMes.levelIndex;
                        servMes.numParticle = 0;
                        servMes.flag_send_bins = 0;

                        servMes.messageSize = servMes.byteSize();
                        pts.sendMessage( servMes );
                    }
                    else if ( clntMes.timeParam == 2 )
                    {

                        strncpy( servMes.header, "JPTP /1.0 100 OK\r\n", 18 );
                        servMes.messageSize = servMes.byteSize();
                        servMes.timeStep = clntMes.step;
                        servMes.levelIndex = clntMes.levelIndex;
                        servMes.repeatLevel = clntMes.repeatLevel;
                        param.sampling_method = clntMes.samplingMethod;
                        param.component_Id = clntMes.renderingId;
                        param.crop.set_enable( clntMes.enable_crop_region );
                        param.crop.set( clntMes.crop_region );
                        param.particle_limit = clntMes.particle_limit;
                        param.particle_density = clntMes.particle_density;
#ifdef CPU_VER
                        transfunc_creator.setProtocol( clntMes );
                        transfunc_creator.setAsisTransferFunction( param.transfer_function );
                        param.transfunc_synthesizer = transfunc_creator.create();
                        param.transfunc_synthesizer->optimize2();
#endif

                        if ( clntMes.nodeType == 'a' )
                        {
                            useAllNodes = true;
                        }
                        else if ( clntMes.nodeType == 's' )
                        {
                            useAllNodes = false;
                        }
                        else
                        {
                            assert( false );
                        }
                        if ( param.GT5D == true || param.GT5D_full == true )
                        {
                            int timeStep = servMes.timeStep;

                            if ( servMes.timeStep > 1 )
                            {
                                for ( int nf = 0; nf < point_creator_lst.size(); nf++ )
                                    point_creator_lst[nf].progressValues();
                            }
                        }

                        if ( !param.hasOption( "L" ) ) param.latency_threshold = -1.0;

                        if ( param.crop.isenabled() )
                        {
                            jd.Initialize( clntMes.step, clntMes.step, fil.total_numSubVolumes,
                                           fil.total_minSubVolumeCoord,
                                           fil.total_maxSubVolumeCoord,
                                           param.latency_threshold, param.jid_pack_size,
                                           param.crop.get_min_coord(),
                                           param.crop.get_max_coord() );
                            servMes.numVolDiv = jd.GetCountVolumes();
                        }
                        else
                        {
                            jd.Initialize( clntMes.step, clntMes.step, fil.total_numSubVolumes,
                                           fil.total_minSubVolumeCoord,
                                           fil.total_maxSubVolumeCoord,
                                           param.latency_threshold, param.jid_pack_size );
                            servMes.numVolDiv = fil.total_numSubVolumes;
                        }

                        if ( timer_count <= TIMER_COUNT_NUM )
                        {
                            TIMER_STA( 470 );
                        }

                        param.sampling_step = CalculateSamplingStep( fil );
                        param.subpixel_level = CalculateSubpixelLevel( param, fil, camera );

                        VariableRange vr;

                        pts.sendMessage( servMes );

                        servMes.tf_count = TF_COUNT;
                        servMes.c_nbins = new kvs::UInt64[clntMes.transfunc.size()];
                        servMes.o_nbins = new kvs::UInt64[clntMes.transfunc.size()];

                        servMes.c_bins.resize( clntMes.transfunc.size() );
                        servMes.o_bins.resize( clntMes.transfunc.size() );

                        c_bins_size = 0;
                        o_bins_size = 0;

                        for ( int tf = 0; tf < servMes.tf_count; tf++ )
                        {
                            servMes.c_nbins[tf] = DEFAULT_NBINS;
                            servMes.o_nbins[tf] = DEFAULT_NBINS;
                            servMes.c_bins[tf] =  new kvs::UInt64[ servMes.c_nbins[tf] ];
                            servMes.o_bins[tf] =  new kvs::UInt64[ servMes.o_nbins[tf] ];
                            c_bins_size += servMes.c_nbins[tf];
                            o_bins_size += servMes.o_nbins[tf];
                            for ( int res = 0; res < servMes.c_nbins[tf]; res++ )
                            {
                                servMes.c_bins[tf][res] = 0;
                            }
                            for ( int res = 0; res < servMes.o_nbins[tf]; res++ )
                            {
                                servMes.o_bins[tf][res] = 0;
                            }
                        }

                        tmp_c_bins = new kvs::UInt64[c_bins_size];
                        tmp_o_bins = new kvs::UInt64[o_bins_size];

                        for ( int tf = 0; tf < c_bins_size; tf++ )
                        {
                            tmp_c_bins[tf] = 0;
                        }

                        for ( int tf = 0; tf < o_bins_size; tf++ )
                        {
                            tmp_o_bins[tf] = 0;
                        }

                        while ( jd.DispatchNext( wid, &st, &vl ) )
                        {
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_STA( 471 );
                            }

                            pbvr::PointObject* originalObject = new pbvr::PointObject;
#ifdef CPU_VER
                            int xvl, fidx;
                            fidx = fil.GetFileIdx( vl, xvl );
                            FilterInfo& fi = fil.m_list[fidx];

                            pbvr::PointObject* tmp_obj = NULL;
                            std::stringstream suffix;
                            suffix << '_' << std::setw( 5 ) << std::setfill( '0' ) << ( st )
                                   << '_' << std::setw( 7 ) << std::setfill( '0' ) << ( xvl + 1 )
                                   << '_' << std::setw( 7 ) << std::setfill( '0' ) << fi.numSubVolumes;
                            kvs::File ifpx( fil.m_list[fidx].filePath );
                            param.input_data = ifpx.pathName() + ifpx.Separator()
                                               + ifpx.baseName() + suffix.str() + ".kvsml";
                            int timeStep = 1;
                            if ( fi.fileType == 1 || fi.fileType == 2 ) // filetype: gathered subvolume or gathered timestep
                            {
                                tmp_obj = point_creator_lst[fidx].run( param, *clntMes.camera, timeStep, st, xvl );

                            }
                            else     // filetype: kvsml
                            {
                                tmp_obj = point_creator_lst[fidx].run( param, *clntMes.camera, timeStep );
                            }

                            size_t nmemb = tmp_obj->nvertices() * 3;
                            kvs::ValueArray<kvs::Real32> coords_array ( tmp_obj->coords().pointer(), nmemb );
                            kvs::ValueArray<kvs::UInt8>  colors_array ( tmp_obj->colors().pointer(), nmemb );
                            kvs::ValueArray<kvs::Real32> normals_array( tmp_obj->normals().pointer(), nmemb );

                            originalObject->clear();
                            originalObject->setCoords( coords_array );
                            originalObject->setColors( colors_array );
                            originalObject->setNormals( normals_array );

                            vr = param.transfunc_synthesizer->variableRange();

                            for ( int tf = 0; tf < tmp_obj->getColorHistogram().size(); tf++ )
                            {
                                servMes.c_nbins[tf] = tmp_obj->getColorHistogram()[ tf ].nbins();
                                for ( int res = 0; res < servMes.c_nbins[tf]; res++ )
                                {
                                    servMes.c_bins[tf][res] += static_cast<kvs::UInt64>( tmp_obj->getColorHistogram()[ tf ][res] );
                                }
                            }

                            for ( int tf = 0; tf < tmp_obj->getOpacityHistogram().size(); tf++ )
                            {
                                servMes.o_nbins[tf] = tmp_obj->getOpacityHistogram()[ tf ].nbins();
                                for ( int res = 0; res < servMes.o_nbins[tf]; res++ )
                                {
                                    servMes.o_bins[tf][res] += static_cast<kvs::UInt64>( tmp_obj->getOpacityHistogram()[ tf ][res] );
                                }
                            }
#else
                            jc.JobCollect( originalObject, &vr, &wid );
#endif
                            int nvertices = originalObject->coords().size() / 3;

                            pbvr::PointObject* object = originalObject;

                            if ( originalObject != object ) delete originalObject;
                            servMes.numParticle = object->coords().size() / 3;
                            if ( servMes.numParticle > 0 )
                            {
                                servMes.positions = new float[3 * servMes.numParticle];
                                servMes.normals = new float[3 * servMes.numParticle];
                                servMes.colors = new unsigned char[3 * servMes.numParticle];
                            }
                            else
                            {
                                servMes.positions = NULL;
                                servMes.normals   = NULL;
                                servMes.colors    = NULL;
                            }
                            for ( int i = 0; i < servMes.numParticle; ++i )
                            {
                                servMes.positions[3 * i + 0] = object->coords()[3 * i + 0];
                                servMes.positions[3 * i + 1] = object->coords()[3 * i + 1];
                                servMes.positions[3 * i + 2] = object->coords()[3 * i + 2];
                                servMes.normals[3 * i + 0] = object->normals()[3 * i + 0];
                                servMes.normals[3 * i + 1] = object->normals()[3 * i + 1];
                                servMes.normals[3 * i + 2] = object->normals()[3 * i + 2];
                                servMes.colors[3 * i + 0] = object->colors()[3 * i + 0];
                                servMes.colors[3 * i + 1] = object->colors()[3 * i + 1];
                                servMes.colors[3 * i + 2] = object->colors()[3 * i + 2];
                            }
                            servMes.varRange = vr;
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_END( 471 );
                            }
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_STA( 472 );
                            }
                            servMes.messageSize = servMes.byteSize();
                            pts.sendMessage( servMes );
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_END( 472 );
                            }
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_STA( 473 );
                            }
                            delete[] servMes.positions;
                            delete[] servMes.normals;
                            delete[] servMes.colors;
                            delete object;
                            if ( timer_count <= TIMER_COUNT_NUM )
                            {
                                TIMER_END( 473 );
                            }
                        } // end of while(DispatchNext)
#ifndef CPU_VER
                        MPI_Allreduce( MPI_IN_PLACE, tmp_c_bins, c_bins_size, MPI_UNSIGNED_LONG, MPI_SUM , MPI_COMM_WORLD );
                        MPI_Allreduce( MPI_IN_PLACE, tmp_o_bins, o_bins_size, MPI_UNSIGNED_LONG, MPI_SUM , MPI_COMM_WORLD );

                        int c_count = 0;
                        int o_count = 0;
                        for ( int tf = 0; tf < servMes.tf_count ; tf++ )
                        {
                            for ( int res = 0; res < servMes.c_nbins[tf]; res++ )
                            {
                                servMes.c_bins[tf][res] = tmp_c_bins[c_count];
                                c_count++;
                            }

                            for ( int res = 0; res < servMes.o_nbins[tf]; res++ )
                            {
                                servMes.o_bins[tf][res] = tmp_o_bins[o_count];
                                o_count++;
                            }
                        }
#endif
                        servMes.flag_send_bins = 1;
                        servMes.subPixelLevel = param.subpixel_level;
                        servMes.messageSize = servMes.byteSize();
                        pts.sendMessage( servMes );

                        for ( int tf = 0; tf < servMes.tf_count; tf++ )
                        {
                            delete[] servMes.c_bins[tf];
                            delete[] servMes.o_bins[tf];
                        }
                        delete[] servMes.c_nbins;
                        delete[] servMes.o_nbins;
                        servMes.tf_count = 0;
                        servMes.flag_send_bins = 0;
                        delete[] tmp_c_bins;
                        delete[] tmp_o_bins;

                        delete param.transfunc_synthesizer;

                        if ( timer_count <= TIMER_COUNT_NUM )
                        {
                            TIMER_END( 470 );
                        }
                    } // end of timeParam == 2
                    else
                    {
                        break;
                    }
                    if ( timer_count <= TIMER_COUNT_NUM )
                    {
                        TIMER_END( 461 );
                    }
                    if ( timer_count == TIMER_COUNT_NUM )
                    {
                        TIMER_END( 1 );
                        TIMER_FIN();
                    }
                } // end of initParam >= 0
            } // end of while (pts.good)

            delete clntMes.camera;

            bsz = -1;
#ifndef CPU_VER
            MPI_Bcast( &bsz, 1, MPI_INT, 0, MPI_COMM_WORLD ); // termination message
#endif

            pts.termServer();
        }
    }
#ifdef KMATH
    km_random.finalize();
#endif
    if ( param.batch == true )
    {
        TIMER_END( 1 );
        TIMER_FIN();
    }
#ifndef CPU_VER
    MPI::Finalize();
#endif
    return retval;
}
